import collections

import th_helpers.utils.card_cache


ENERGY = {
    'Grass': 'G',
    'Fire': 'R',
    'Water': 'W',
    'Lightning': 'L',
    'Psychic': 'P',
    'Fighting': 'F',
    'Darkness': 'D',
    'Metal': 'M',
    'Fairy': 'Y',
    'Dragon': 'N',
    'Colorless': 'C'
}

card_cache = {}


def _query_card(id):
    card = th_helpers.utils.card_cache.get_card(id)
    return card


def get_card(name, set, number):
    card = {'name': name, 'set': set, 'number': number}
    set_obj = th_helpers.utils.card_cache.get_set_by_ptcgo(set)
    if set_obj is None:
        return card
    card_set = set_obj['id']
    card_number = card['number']

    obj = _query_card(f'{card_set}-{card_number}')
    if obj is None:
        basic_energy = not any(c.isdigit() for c in card['number']) or ('Energy' in card['name'] and card['name'].split(' ')[0] in ENERGY)
        card['supertype'] = 'Energy' if basic_energy else None
        card['subtype'] = 'Basic' if basic_energy else None
        card['dex'] = None
    else:
        card['supertype'] = obj.get('supertype')
        subtypes = obj.get('subtypes', [])
        card['subtype'] = subtypes[0] if 'Pokémon Tool' not in subtypes else 'Pokémon Tool'
        nat_dex = obj.get('nationalPokedexNumbers')
        card['dex'] = nat_dex[0] if nat_dex else None

    if card['supertype'] == 'Pokémon':
        card_id = f"{card['name']}-{card['set']}-{card['number']}"
        if card_id not in card_cache:
            card['attacks'] = obj.get('attacks', [])
            # TODO check similar cards in cache
            possible_matches = {k: v for k, v in card_cache.items() if k.startswith(card['name'])}
            matched = False
            for k, v in possible_matches.items():
                if v.get('attacks', []) == card['attacks']:
                    matched = True
                    card_cache[card_id] = v
                    break
            if not matched:
                card_cache[card_id] = card
        return card_cache[card_id]
    else:
        if name not in card_cache:
            card_cache[name] = card
        return card_cache[name]


def sort_pokemon(cards):
    if len(cards) == 0:
        return []
    name_counter = {}
    name_dex = {}
    for card in cards:
        card_name = card['name']
        if card_name not in name_dex:
            name_counter[card_name] = 0
            name_dex[card_name] = 0
        name_counter[card_name] += card.get('count', 0)
        if card['dex']:
            name_dex[card_name] = card['dex']

    sort = []
    while len(name_counter.keys()) > 0:
        max_value = max(name_counter.values())
        max_names = [k for k, v in name_counter.items() if v == max_value]
        if max_value <= 1:
            remaining = set(name_counter.keys())
            sort.extend(sorted(
                (c for c in cards if c['name'] in remaining),
                key=lambda x: x.get('play_rate', 0),
                reverse=True,
            ))
            break
        min_dex = min(v for k, v in name_dex.items() if k in max_names)
        min_name = next(k for k, v in name_dex.items() if v == min_dex and k in max_names)

        named_cards = (c for c in cards if c['name'] == min_name)
        sort.extend(sorted(named_cards, key=lambda x: (x.get('count', 0), x.get('play_rate', 0)), reverse=True))
        del name_counter[min_name]
        evolution_cards = [c for c in cards if (name_dex[c['name']] == min_dex + 1 or name_dex[c['name']] == min_dex + 2) and c['subtype'] in ['Stage 1', 'Stage 2'] and c['name'] in name_counter]
        for evo in evolution_cards:
            if evo['name'] in name_counter:
                del name_counter[evo['name']]

        sort.extend(sorted(evolution_cards, key=lambda x: (name_dex[x['name']], 60-x.get('count', 0), 1-x.get('play_rate', 0))))
    return sort


def sort_trainers(cards):
    if len(cards) == 0:
        return []
    custom_order = {'Supporter': 1, 'Item': 2, 'Pokémon Tool': 3, 'Stadium': 4}
    return sorted(cards, key=lambda x: (custom_order[x['subtype']], 60-x.get('count', 0), 1-x.get('play_rate', 0)))


def sort_energy(cards):
    if len(cards) > 0 and 'count' in cards[0]:
        return sorted(cards, key=lambda x: (x.get('count', 0), x.get('play_rate', 0)), reverse=True)
    return cards


def sort_deck(cards):
    '''
    Sorting requirements:
    - Pokemon
        - Count
    - Trainer
    - Energy
    '''
    pokemon = 'Pokémon'
    trainer = 'Trainer'
    energy = 'Energy'
    types = collections.OrderedDict({
        pokemon: [],
        trainer: [],
        energy: []
    })
    sort_functions = {
        pokemon: sort_pokemon,
        trainer: sort_trainers,
        energy: sort_energy
    }
    for c in cards:
        supertype = c['supertype']
        if supertype not in types:
            supertype = energy
        types[supertype].append(c)
    sorted_cards = []
    for t in types.keys():
        sorted_cards.extend(
            sort_functions[t](types[t])
        )
    return sorted_cards


if __name__ == '__main__':
    c = {'set': 'SVI', 'number': '34'}
    c = get_card(c)
    print(c)

    d = {'set': 'BRS', 'number': 'R'}
    d = get_card(d)
    print(d)



import th_helpers.scraper.utils as utils

def prompt_to_get_pokedata_tour_id():
    print('* Finding Tournaments on PokeData')
    base_url = 'https://pokedata.ovh/standings/'
    page_html = utils.get_html(base_url)
    buttons = page_html.findAll('button')
    ids = {}
    for i, button in enumerate(buttons[:10]):
        ids[i] = button['onclick'].split('=')[1][1:-2]
        name = button.text.replace("\n", "")
        print(f'[{i}] {name}')
    selected_id = input('Select tournament: ')
    print('')
    return ids[int(selected_id)]



# card_cache.py
from __future__ import annotations

import io
import json
import os
import shutil
import tempfile
import threading
import zipfile
from typing import Any, Dict, List, Optional, Sequence

import requests

# ---------------- Config ----------------
CACHE_DIR = os.environ.get("POKEMON_CACHE_DIR", os.path.join(os.getcwd(), "data"))
EXTRACT_DIR = os.path.join(CACHE_DIR, "pokemon-tcg-data")   # extracted archieve

# Cards index (existing behavior)
INDEX_PATH = os.path.join(CACHE_DIR, "cards_by_id.json")    # flat {card_id: card_json}
# Sets index (new)
SETS_INDEX_PATH = os.path.join(CACHE_DIR, "sets_by_ptcgo.json")  # flat {PTCGO_CODE: set_json}

GH_OWNER = os.environ.get("POKEMON_GH_OWNER", "PokemonTCG")
GH_REPO = os.environ.get("POKEMON_GH_REPO", "pokemon-tcg-data")
GH_TOKEN = os.environ.get("GITHUB_TOKEN")  # optional for higher rate limits
GH_BRANCH = os.environ.get("POKEMON_GH_BRANCH", "master")

RELEASE_TAG = os.environ.get("POKEMON_RELEASE_TAG")        # e.g. "v2.15"
FORCE_REFRESH = os.environ.get("POKEMON_FORCE_REFRESH", "0") == "1"

# ---------------- State -----------------
_lock = threading.RLock()
_loaded = False

_cards_by_id: Dict[str, Dict[str, Any]] = {}
_sets_by_ptcgo: Dict[str, Dict[str, Any]] = {}  # NEW


# ---------------- GitHub helpers --------
def _gh_headers() -> Dict[str, str]:
    h = {"Accept": "application/vnd.github+json", "User-Agent": "pokemon-tcg-data-client"}
    if GH_TOKEN:
        h["Authorization"] = f"Bearer {GH_TOKEN}"
    return h


def _tag_release_json(tag: str) -> Dict[str, Any]:
    url = f"https://api.github.com/repos/{GH_OWNER}/{GH_REPO}/releases/tags/{tag}"
    r = requests.get(url, headers=_gh_headers(), timeout=30)
    r.raise_for_status()
    return r.json()


def _pick_zip_asset(rel: Dict[str, Any]) -> str:
    # Prefer a .zip asset; fall back to source zipball
    for a in rel.get("assets", []):
        name = (a.get("name") or "").lower()
        if name.endswith(".zip") and a.get("browser_download_url"):
            return a["browser_download_url"]
    if rel.get("zipball_url"):
        return rel["zipball_url"]
    raise RuntimeError("No downloadable zip asset found for this release.")


# ---------------- Disk helpers ----------
def _atomic_write_text(path: str, text: str) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    fd, tmp = tempfile.mkstemp(prefix=".tmp-", dir=os.path.dirname(path))
    try:
        with os.fdopen(fd, "w", encoding="utf-8") as f:
            f.write(text)
        os.replace(tmp, path)
    finally:
        try:
            if os.path.exists(tmp):
                os.remove(tmp)
        except OSError:
            pass


def _load_cards_index_from_disk() -> bool:
    if not os.path.isfile(INDEX_PATH):
        return False
    try:
        with open(INDEX_PATH, "r", encoding="utf-8") as f:
            data = json.load(f)
        if isinstance(data, dict):
            _cards_by_id.clear()
            _cards_by_id.update(data)
            return True
    except Exception:
        return False
    return False


def _save_cards_index_to_disk() -> None:
    _atomic_write_text(INDEX_PATH, json.dumps(_cards_by_id, ensure_ascii=False))


def _load_sets_index_from_disk() -> bool:
    if not os.path.isfile(SETS_INDEX_PATH):
        return False
    try:
        with open(SETS_INDEX_PATH, "r", encoding="utf-8") as f:
            data = json.load(f)
        if isinstance(data, dict):
            _sets_by_ptcgo.clear()
            _sets_by_ptcgo.update(data)
            return True
    except Exception:
        return False
    return False


def _save_sets_index_to_disk() -> None:
    _atomic_write_text(SETS_INDEX_PATH, json.dumps(_sets_by_ptcgo, ensure_ascii=False))


# -------------- Extract & index ---------
def _extract_zip_to_dir(zip_bytes: bytes, target_dir: str) -> None:
    if os.path.isdir(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir, exist_ok=True)
    with zipfile.ZipFile(io.BytesIO(zip_bytes)) as zf:
        zf.extractall(target_dir)


def _iter_json_files(root: str) -> List[str]:
    paths: List[str] = []
    for base, _, files in os.walk(root):
        for fn in files:
            if fn.lower().endswith(".json"):
                paths.append(os.path.join(base, fn))
    return paths


def _index_cards_from_extracted(root: str) -> Dict[str, Dict[str, Any]]:
    """
    Build a FLAT {card_id: card_json}. Handles these shapes:
      - [ {...card...}, {...card...} ]
      - { "data": [ {...card...} ] }
      - { "data": { "cards": [ {...card...} ] } }   # some set wrappers
      - { ...single-card... }  (dict with 'id')
    Ignores JSON that doesn't represent cards.
    """
    by_id: Dict[str, Dict[str, Any]] = {}

    def add_card(maybe: Any):
        if isinstance(maybe, dict) and "id" in maybe:
            by_id[str(maybe["id"])] = maybe

    for path in _iter_json_files(root):
        p_norm = path.replace("\\", "/")
        if "/cards/en/" not in p_norm:
            continue
        try:
            with open(path, "r", encoding="utf-8") as f:
                obj = json.load(f)
        except Exception:
            continue

        if isinstance(obj, list):
            for item in obj:
                add_card(item)
            continue

        if isinstance(obj, dict):
            if "data" in obj:
                data = obj["data"]
                if isinstance(data, list):
                    for item in data:
                        add_card(item)
                    continue
                if isinstance(data, dict) and "cards" in data and isinstance(data["cards"], list):
                    for item in data["cards"]:
                        add_card(item)
                    continue
            if "cards" in obj and isinstance(obj["cards"], list):
                for item in obj["cards"]:
                    add_card(item)
                continue
            add_card(obj)  # last chance: single-card dict

    return by_id


def _index_sets_from_extracted(root: str) -> Dict[str, Dict[str, Any]]:
    """
    Build a FLAT {PTCGO_CODE: set_json} from /sets/en.json inside the extracted repo.
    Expected shapes:
      - [ {...set...}, ... ]
      - { "data": [ {...set...}, ... ] }
    Only entries with a non-empty 'ptcgoCode' are included.
    """
    # Find the sets/en.json file anywhere under the extracted root
    sets_path = None
    for path in _iter_json_files(root):
        p_norm = path.replace("\\", "/")
        if p_norm.endswith("/sets/en.json"):
            sets_path = path
            break
    if not sets_path:
        return {}

    try:
        with open(sets_path, "r", encoding="utf-8") as f:
            obj = json.load(f)
    except Exception:
        return {}

    raw_list: List[Dict[str, Any]] = []
    if isinstance(obj, list):
        raw_list = [x for x in obj if isinstance(x, dict)]
    elif isinstance(obj, dict) and "data" in obj and isinstance(obj["data"], list):
        raw_list = [x for x in obj["data"] if isinstance(x, dict)]

    by_code: Dict[str, Dict[str, Any]] = {}
    for s in raw_list:
        code = s.get("ptcgoCode")
        if isinstance(code, str) and code.strip():
            by_code[code.strip().upper()] = s
    return by_code


# ---------------- Public API ------------
def ensure_loaded(force_refresh: bool = FORCE_REFRESH) -> None:
    """
    Ensure in-memory flat indexes:
      - Cards: {card_id: card_json}
      - Sets:  {PTCGO_CODE: set_json}
    If missing (or forced), fetch the GitHub archive (tagged release when specified, otherwise
    the configured branch) and rebuild both.
    """
    global _loaded
    with _lock:
        if _loaded and not force_refresh:
            return

        if not force_refresh:
            cards_ok = _load_cards_index_from_disk()
            sets_ok = _load_sets_index_from_disk()
            if cards_ok and sets_ok:
                _loaded = True
                return

        # (Re)build from a fresh archive
        if RELEASE_TAG:
            rel = _tag_release_json(RELEASE_TAG)
            zip_url = _pick_zip_asset(rel)
            r = requests.get(zip_url, headers=_gh_headers(), timeout=120)
            r.raise_for_status()
            zip_bytes = r.content
        else:
            branch = GH_BRANCH or "master"
            url = f"https://api.github.com/repos/{GH_OWNER}/{GH_REPO}/zipball/{branch}"
            r = requests.get(url, headers=_gh_headers(), timeout=120)
            r.raise_for_status()
            zip_bytes = r.content

        _extract_zip_to_dir(zip_bytes, EXTRACT_DIR)
        flat_cards = _index_cards_from_extracted(EXTRACT_DIR)
        flat_sets = _index_sets_from_extracted(EXTRACT_DIR)

        _cards_by_id.clear()
        _cards_by_id.update(flat_cards)
        _save_cards_index_to_disk()

        _sets_by_ptcgo.clear()
        _sets_by_ptcgo.update(flat_sets)
        _save_sets_index_to_disk()

        _loaded = True


def update_from_release(tag: Optional[str] = None) -> int:
    """
    Force-refresh from a specific tag (e.g., 'v2.15') or the configured branch if None.
    Returns the number of cards indexed (for backward compatibility).
    Note: sets are also refreshed as part of this call.
    """
    global RELEASE_TAG
    if tag:
        RELEASE_TAG = tag
    ensure_loaded(force_refresh=True)
    return len(_cards_by_id)


# ---- Cards lookups (existing API) ----
def get_card(card_id: str) -> Optional[Dict[str, Any]]:
    ensure_loaded()
    return _cards_by_id.get(str(card_id))


def get_total_length(card_id: str) -> Optional[Dict[str, Any]]:
    ensure_loaded()
    return len(_cards_by_id.keys())


def get_cards_by_ids(ids: Sequence[str]) -> Dict[str, Dict[str, Any]]:
    ensure_loaded()
    return {str(cid): _cards_by_id[str(cid)] for cid in ids if str(cid) in _cards_by_id}


# ---- Sets lookups ----
def get_set_by_ptcgo(code: str) -> Optional[Dict[str, Any]]:
    """
    Fetch a set object by its PTCGO code (case-insensitive).
    """
    ensure_loaded()
    if code is None:
        return None
    # TODO build out this mapping better
    if code == 'SVP':
        code = 'PR-SV'
    return _sets_by_ptcgo.get(str(code).strip().upper())


def get_sets_by_ptcgocodes(codes: Sequence[str]) -> Dict[str, Dict[str, Any]]:
    """
    Batch fetch set objects by PTCGO codes (case-insensitive).
    """
    ensure_loaded()
    out: Dict[str, Dict[str, Any]] = {}
    for c in codes:
        k = str(c).strip().upper()
        if k in _sets_by_ptcgo:
            out[k] = _sets_by_ptcgo[k]
    return out


def get_all_sets_map() -> Dict[str, Dict[str, Any]]:
    """
    Return the full {PTCGO_CODE: set_json} mapping.
    """
    ensure_loaded()
    return dict(_sets_by_ptcgo)


# ---------------- Local testing ----------
if __name__ == "__main__":
    import argparse

    ap = argparse.ArgumentParser(description="Fetch & index Pokémon TCG data (flat by card_id and PTCGO sets).")
    ap.add_argument(
        "--tag",
        default=None,
        help="Release tag to fetch (e.g., v2.15). Defaults to the configured branch when omitted.",
    )
    ap.add_argument("--force", action="store_true", help="Force refresh even if cached index exists.")
    ap.add_argument("--ids", nargs="*", default=[], help="Card IDs to print after loading.")
    ap.add_argument("--set-codes", nargs="*", default=[], help="PTCGO set codes to print after loading.")
    args = ap.parse_args()

    if args.force:
        FORCE_REFRESH = True

    n_cards = update_from_release(args.tag) if (args.force or args.tag) else (ensure_loaded() or len(_cards_by_id))
    print(f"Loaded {len(_cards_by_id)} cards and {len(_sets_by_ptcgo)} sets.")

    ids = args.ids or list(_cards_by_id.keys())[:5]
    for cid, card in get_cards_by_ids(ids).items():
        print(f"{cid}: {card}")

    if args.set_codes:
        sets = get_sets_by_ptcgocodes(args.set_codes)
        for code, s in sets.items():
            print(f"{code}: {s.get('name')} (id={s.get('id')})")



from dash import html
import dash_bootstrap_components as dbc

DEFAULT_OPTIONS = [
    {'label': html.I(className='fas fa-xmark text-danger'), 'value': -1},
    {'label': html.I(className='far fa-circle'), 'value': 0},
    {'label': html.I(className='fas fa-check text-success'), 'value': 1},
]

TURN_OPTIONS = [
    {'label': html.I(className='fas fa-1 text-info'), 'value': 1},
    {'label': html.I(className='far fa-circle'), 'value': 0},
    {'label': html.I(className='fas fa-2 text-warning'), 'value': 2},
]

def create_ternary_switch(id, label, options=DEFAULT_OPTIONS):
    return html.Div([
        dbc.RadioItems(
            id=id,
            className='btn-group btn-group-sm me-1',
            inputClassName='btn-check',
            labelClassName='btn btn-outline-primary ternary-switch',
            labelCheckedClassName='active',
            options=options,
            value=0,
        ),
        dbc.Label(label)
    ], className='radio-group')



from dash import html
import dash_bootstrap_components as dbc


def create_help_icon(id, children, className='', big=False):
    return html.Div([
        html.I(className='far fa-circle-question', id=id),
        dbc.Popover(children, body=True, target=id, trigger='click hover', placement='bottom', class_name=f'help-icon-popover {"big" if big else ""}')
    ], className=f'd-inline-block {className}')



pokemon = [
    'abomasnow-mega',
    'abomasnow',
    'abra',
    'absol-mega',
    'absol',
    'accelgor',
    'aegislash-blade',
    'aegislash',
    'aerodactyl-mega',
    'aerodactyl',
    'aggron-mega',
    'aggron',
    'aipom',
    'alakazam-mega',
    'alakazam',
    'alcremie-gmax',
    'alcremie',
    'alomomola',
    'altaria-mega',
    'altaria',
    'amaura',
    'ambipom',
    'amoonguss',
    'ampharos-mega',
    'ampharos',
    'annihilape',
    'anorith',
    'appletun-gmax',
    'appletun',
    'applin',
    'araquanid',
    'arbok',
    'arboliva',
    'arcanine-hisui',
    'arcanine',
    'arceus',
    'archaludon',
    'archen',
    'archeops',
    'arctibax',
    'arctovish',
    'arctozolt',
    'ariados',
    'armaldo',
    'armarouge',
    'aromatisse',
    'aron',
    'arrokuda',
    'articuno-galar',
    'articuno',
    'audino-mega',
    'audino',
    'aurorus',
    'avalugg-hisui',
    'avalugg',
    'axew',
    'azelf',
    'azumarill',
    'azurill',
    'bagon',
    'baltoy',
    'banette-mega',
    'banette',
    'barbaracle',
    'barboach',
    'barraskewda',
    'basculegion',
    'basculin',
    'bastiodon',
    'baxcalibur',
    'bayleef',
    'beartic',
    'beautifly',
    'beedrill-mega',
    'beedrill',
    'beheeyem',
    'beldum',
    'bellibolt',
    'bellossom',
    'bellsprout',
    'bergmite',
    'bewear',
    'bibarel',
    'bidoof',
    'binacle',
    'bisharp',
    'blacephalon',
    'blastoise-gmax',
    'blastoise-mega',
    'blastoise',
    'blaziken-mega',
    'blaziken',
    'blipbug',
    'blissey',
    'blitzle',
    'boldore',
    'boltund',
    'bombirdier',
    'bonsly',
    'bouffalant',
    'bounsweet',
    'braixen',
    'brambleghast',
    'bramblin',
    'braviary-hisui',
    'braviary',
    'breloom',
    'brionne',
    'bronzong',
    'bronzor',
    'brute-bonnet',
    'bruxish',
    'budew',
    'buizel',
    'bulbasaur',
    'buneary',
    'bunnelby',
    'burmy',
    'butterfree-gmax',
    'butterfree',
    'buzzwole',
    'cacnea',
    'cacturne',
    'calyrex-ice-rider',
    'calyrex',
    'calyrex-shadow-rider',
    'camerupt-mega',
    'camerupt',
    'capsakid',
    'carbink',
    'carkol',
    'carnivine',
    'carracosta',
    'carvanha',
    'cascoon',
    'castform',
    'castform-rainy',
    'castform-snowy',
    'castform-sunny',
    'caterpie',
    'celebi',
    'celesteela',
    'centiskorch-gmax',
    'centiskorch',
    'ceruledge',
    'cetitan',
    'cetoddle',
    'chandelure',
    'chansey',
    'charcadet',
    'charizard-gmax',
    'charizard-mega-x',
    'charizard-mega-y',
    'charizard',
    'charjabug',
    'charmander',
    'charmeleon',
    'chatot',
    'cherrim',
    'cherrim-sunshine',
    'cherubi',
    'chesnaught',
    'chespin',
    'chewtle',
    'chien-pao',
    'chikorita',
    'chimchar',
    'chimecho',
    'chinchou',
    'chingling',
    'chi-yu',
    'cinccino',
    'cinderace-gmax',
    'cinderace',
    'clamperl',
    'clauncher',
    'clawitzer',
    'claydol',
    'clefable',
    'clefairy',
    'cleffa',
    'clobbopus',
    'clodsire',
    'cloyster',
    'coalossal-gmax',
    'coalossal',
    'cobalion',
    'cofagrigus',
    'combee',
    'combusken',
    'comfey',
    'conkeldurr',
    'copperajah-gmax',
    'copperajah',
    'corphish',
    'corsola-galar',
    'corsola',
    'corviknight-gmax',
    'corviknight',
    'corvisquire',
    'cosmoem',
    'cosmog',
    'cottonee',
    'crabominable',
    'crabrawler',
    'cradily',
    'cramorant',
    'cranidos',
    'crawdaunt',
    'cresselia',
    'croagunk',
    'crobat',
    'crocalor',
    'croconaw',
    'crustle',
    'cryogonal',
    'cubchoo',
    'cubone',
    'cufant',
    'cursola',
    'cutiefly',
    'cyclizar',
    'cyndaquil',
    'dachsbun',
    'darkrai',
    'darmanitan-galar',
    'darmanitan-galar-zen',
    'darmanitan',
    'darmanitan-zen',
    'dartrix',
    'darumaka-galar',
    'darumaka',
    'decidueye-hisui',
    'decidueye',
    'dedenne',
    'deerling',
    'deino',
    'delcatty',
    'delibird',
    'delphox',
    'deoxys-attack',
    'deoxys-defense',
    'deoxys',
    'deoxys-speed',
    'dewgong',
    'dewott',
    'dewpider',
    'dhelmise',
    'dialga-origin',
    'dialga',
    'diancie-mega',
    'diancie',
    'diggersby',
    'diglett-alola',
    'diglett',
    'dipplin',
    'ditto',
    'dodrio',
    'doduo',
    'dolliv',
    'dondozo',
    'donphan',
    'dottler',
    'doublade',
    'dracovish',
    'dracozolt',
    'dragalge',
    'dragapult',
    'dragonair',
    'dragonite',
    'drakloak',
    'drampa',
    'drapion',
    'dratini',
    'drednaw-gmax',
    'drednaw',
    'dreepy',
    'drifblim',
    'drifloon',
    'drilbur',
    'drizzile',
    'drowzee',
    'druddigon',
    'dubwool',
    'ducklett',
    'dudunsparce',
    'dugtrio-alola',
    'dugtrio',
    'dunsparce',
    'duosion',
    'duraludon-gmax',
    'duraludon',
    'durant',
    'dusclops',
    'dusknoir',
    'duskull',
    'dustox',
    'dwebble',
    'eelektrik',
    'eelektross',
    'eevee-gmax',
    'eevee',
    'eiscue-noice',
    'eiscue',
    'ekans',
    'eldegoss',
    'electabuzz',
    'electivire',
    'electrike',
    'electrode-hisui',
    'electrode',
    'elekid',
    'elgyem',
    'emboar',
    'emolga',
    'empoleon',
    'enamorus',
    'enamorus-therian',
    'entei',
    'escavalier',
    'espathra',
    'espeon',
    'espurr',
    'eternatus-eternamax',
    'eternatus',
    'excadrill',
    'exeggcute',
    'exeggutor-alola',
    'exeggutor',
    'exploud',
    'falinks',
    'farfetchd-galar',
    'farfetchd',
    'farigiraf',
    'fearow',
    'feebas',
    'fennekin',
    'feraligatr',
    'ferroseed',
    'ferrothorn',
    'fezandipiti',
    'fidough',
    'finizen',
    'finneon',
    'flaaffy',
    'flabebe',
    'flamigo',
    'flapple-gmax',
    'flapple',
    'flareon',
    'fletchinder',
    'fletchling',
    'flittle',
    'floatzel',
    'floette',
    'floragato',
    'florges',
    'flutter-mane',
    'flygon',
    'fomantis',
    'foongus',
    'forretress',
    'fraxure',
    'frigibax',
    'frillish',
    'froakie',
    'frogadier',
    'froslass',
    'frosmoth',
    'fuecoco',
    'furfrou',
    'furret',
    'gabite',
    'gallade-mega',
    'gallade',
    'galvantula',
    'garbodor-gmax',
    'garbodor',
    'garchomp-mega',
    'garchomp',
    'gardevoir-mega',
    'gardevoir',
    'garganacl',
    'gastly',
    'gastrodon',
    'genesect',
    'gengar-gmax',
    'gengar-mega',
    'gengar',
    'geodude-alola',
    'geodude',
    'gholdengo',
    'gible',
    'gigalith',
    'gimmighoul',
    'girafarig',
    'giratina-origin',
    'giratina',
    'glaceon',
    'glalie-mega',
    'glalie',
    'glameow',
    'glastrier',
    'gligar',
    'glimmet',
    'glimmora',
    'gliscor',
    'gloom',
    'gogoat',
    'golbat',
    'goldeen',
    'golduck',
    'golem-alola',
    'golem',
    'golett',
    'golisopod',
    'golurk',
    'goodra-hisui',
    'goodra',
    'goomy',
    'gorebyss',
    'gossifleur',
    'gothita',
    'gothitelle',
    'gothorita',
    'gouging-fire',
    'gourgeist',
    'grafaiai',
    'granbull',
    'grapploct',
    'graveler-alola',
    'graveler',
    'great-tusk',
    'greavard',
    'greedent',
    'greninja-ash',
    'greninja',
    'grimer-alola',
    'grimer',
    'grimmsnarl-gmax',
    'grimmsnarl',
    'grookey',
    'grotle',
    'groudon',
    'groudon-primal',
    'grovyle',
    'growlithe-hisui',
    'growlithe',
    'grubbin',
    'grumpig',
    'gulpin',
    'gumshoos',
    'gurdurr',
    'guzzlord',
    'gyarados-mega',
    'gyarados',
    'hakamo-o',
    'happiny',
    'hariyama',
    'hatenna',
    'hatterene-gmax',
    'hatterene',
    'hattrem',
    'haunter',
    'hawlucha',
    'haxorus',
    'heatmor',
    'heatran',
    'heliolisk',
    'helioptile',
    'heracross-mega',
    'heracross',
    'herdier',
    'hippopotas',
    'hippowdon',
    'hitmonchan',
    'hitmonlee',
    'hitmontop',
    'honchkrow',
    'honedge',
    'ho-oh',
    'hoopa',
    'hoopa-unbound',
    'hoothoot',
    'hoppip',
    'horsea',
    'houndoom-mega',
    'houndoom',
    'houndour',
    'houndstone',
    'huntail',
    'hydrapple',
    'hydreigon',
    'hypno',
    'igglybuff',
    'illumise',
    'impidimp',
    'incineroar',
    'indeedee',
    'infernape',
    'inkay',
    'inteleon-gmax',
    'inteleon',
    'iron-boulder',
    'iron-bundle',
    'iron-crown',
    'iron-hands',
    'iron-jugulis',
    'iron-leaves',
    'iron-moth',
    'iron-thorns',
    'iron-treads',
    'iron-valiant',
    'ivysaur',
    'jangmo-o',
    'jellicent',
    'jigglypuff',
    'jirachi',
    'jolteon',
    'joltik',
    'jumpluff',
    'jynx',
    'kabuto',
    'kabutops',
    'kadabra',
    'kakuna',
    'kangaskhan-mega',
    'kangaskhan',
    'karrablast',
    'kartana',
    'kecleon',
    'keldeo',
    'keldeo-resolute',
    'kilowattrel',
    'kingambit',
    'kingdra',
    'kingler-gmax',
    'kingler',
    'kirlia',
    'klang',
    'klawf',
    'kleavor',
    'klefki',
    'klinklang',
    'klink',
    'koffing',
    'komala',
    'kommo-o',
    'koraidon',
    'krabby',
    'kricketot',
    'kricketune',
    'krokorok',
    'krookodile',
    'kubfu',
    'kyogre',
    'kyogre-primal',
    'kyurem-black',
    'kyurem',
    'kyurem-white',
    'lairon',
    'lampent',
    'landorus',
    'landorus-therian',
    'lanturn',
    'lapras-gmax',
    'lapras',
    'larvesta',
    'larvitar',
    'latias-mega',
    'latias',
    'latios-mega',
    'latios',
    'leafeon',
    'leavanny',
    'lechonk',
    'ledian',
    'ledyba',
    'lickilicky',
    'lickitung',
    'liepard',
    'lileep',
    'lilligant-hisui',
    'lilligant',
    'lillipup',
    'linoone-galar',
    'linoone',
    'litleo',
    'litten',
    'litwick',
    'lokix',
    'lombre',
    'lopunny-mega',
    'lopunny',
    'lotad',
    'loudred',
    'lucario-mega',
    'lucario',
    'ludicolo',
    'lugia',
    'lumineon',
    'lunala',
    'lunatone',
    'lurantis',
    'luvdisc',
    'luxio',
    'luxray',
    'lycanroc-dusk',
    'lycanroc-midnight',
    'lycanroc',
    'mabosstiff',
    'machamp-gmax',
    'machamp',
    'machoke',
    'machop',
    'magby',
    'magcargo',
    'magearna',
    'magikarp',
    'magmar',
    'magmortar',
    'magnemite',
    'magneton',
    'magnezone',
    'makuhita',
    'malamar',
    'mamoswine',
    'manaphy',
    'mandibuzz',
    'manectric-mega',
    'manectric',
    'mankey',
    'mantine',
    'mantyke',
    'maractus',
    'mareanie',
    'mareep',
    'marill',
    'marowak-alola',
    'marowak',
    'marshadow',
    'marshtomp',
    'maschiff',
    'masquerain',
    'maushold',
    'mawile-mega',
    'mawile',
    'medicham-mega',
    'medicham',
    'meditite',
    'meganium',
    'melmetal-gmax',
    'melmetal',
    'meloetta-pirouette',
    'meloetta',
    'meltan',
    'meowscarada',
    'meowstic',
    'meowth-alola',
    'meowth-galar',
    'meowth-gmax',
    'meowth',
    'mesprit',
    'metagross-mega',
    'metagross',
    'metang',
    'metapod',
    'mew',
    'mewtwo-mega-x',
    'mewtwo-mega-y',
    'mewtwo',
    'mienfoo',
    'mienshao',
    'mightyena',
    'milcery',
    'milotic',
    'miltank',
    'mime-jr',
    'mimikyu',
    'minccino',
    'minior',
    'minun',
    'miraidon',
    'misdreavus',
    'mismagius',
    'moltres-galar',
    'moltres',
    'monferno',
    'morelull',
    'morgrem',
    'morpeko-hangry',
    'morpeko',
    'mothim',
    'mr-mime-galar',
    'mr-mime',
    'mr-rime',
    'mudbray',
    'mudkip',
    'mudsdale',
    'muk-alola',
    'muk',
    'munchlax',
    'munkidori',
    'munna',
    'murkrow',
    'musharna',
    'nacli',
    'naclstack',
    'naganadel',
    'natu',
    'necrozma',
    'necrozma-ultra',
    'nickit',
    'nidoking',
    'nidoqueen',
    'nidoran-f',
    'nidoran-m',
    'nidorina',
    'nidorino',
    'nihilego',
    'nincada',
    'ninetales-alola',
    'ninetales',
    'ninjask',
    'noctowl',
    'noibat',
    'noivern',
    'nosepass',
    'numel',
    'nuzleaf',
    'nymble',
    'obstagoon',
    'octillery',
    'oddish',
    'ogerpon-teal-mask',
    'ogerpon-wellspring-mask',
    'ogerpon-hearthflame-mask',
    'ogerpon-cornerstone-mask',
    'oinkologne',
    'okidogi',
    'omanyte',
    'omastar',
    'onix',
    'oranguru',
    'orbeetle-gmax',
    'orbeetle',
    'oricorio-pau',
    'oricorio',
    'oricorio-pom-pom',
    'oricorio-sensu',
    'orthworm',
    'oshawott',
    'overqwil',
    'pachirisu',
    'palafin',
    'palkia-origin',
    'palkia',
    'palossand',
    'palpitoad',
    'pancham',
    'pangoro',
    'panpour',
    'pansage',
    'pansear',
    'parasect',
    'paras',
    'passimian',
    'patrat',
    'pawmi',
    'pawmo',
    'pawmot',
    'pawniard',
    'pecharunt',
    'pelipper',
    'perrserker',
    'persian-alola',
    'persian',
    'petilil',
    'phanpy',
    'phantump',
    'pheromosa',
    'phione',
    'pichu',
    'pidgeot-mega',
    'pidgeot',
    'pidgeotto',
    'pidgey',
    'pidove',
    'pignite',
    'pikachu-gmax',
    'pikachu',
    'pikipek',
    'piloswine',
    'pincurchin',
    'pineco',
    'pinsir-mega',
    'pinsir',
    'piplup',
    'plusle',
    'poipole',
    'politoed',
    'poliwag',
    'poliwhirl',
    'poliwrath',
    'poltchageist',
    'polteageist',
    'ponyta-galar',
    'ponyta',
    'poochyena',
    'popplio',
    'porygon2',
    'porygon',
    'porygon-z',
    'primarina',
    'primeape',
    'prinplup',
    'probopass',
    'psyduck',
    'pumpkaboo',
    'pupitar',
    'purrloin',
    'purugly',
    'pyroar',
    'pyukumuku',
    'quagsire',
    'quaquaval',
    'quaxly',
    'quaxwell',
    'quilava',
    'quilladin',
    'qwilfish-hisui',
    'qwilfish',
    'raboot',
    'rabsca',
    'raging-bolt',
    'raichu-alola',
    'raichu',
    'raikou',
    'ralts',
    'rampardos',
    'rapidash-galar',
    'rapidash',
    'raticate-alola',
    'raticate',
    'rattata-alola',
    'rattata',
    'rayquaza-mega',
    'rayquaza',
    'regice',
    'regidrago',
    'regieleki',
    'regigigas',
    'regirock',
    'registeel',
    'relicanth',
    'rellor',
    'remoraid',
    'reshiram',
    'reuniclus',
    'revavroom',
    'rhydon',
    'rhyhorn',
    'rhyperior',
    'ribombee',
    'rillaboom-gmax',
    'rillaboom',
    'riolu',
    'roaring-moon',
    'rockruff',
    'roggenrola',
    'rolycoly',
    'rookidee',
    'roselia',
    'roserade',
    'rotom-fan',
    'rotom-frost',
    'rotom-heat',
    'rotom-mow',
    'rotom',
    'rotom-wash',
    'rowlet',
    'rufflet',
    'runerigus',
    'sableye-mega',
    'sableye',
    'salamence-mega',
    'salamence',
    'salandit',
    'salazzle',
    'samurott-hisui',
    'samurott',
    'sandaconda-gmax',
    'sandaconda',
    'sandile',
    'sandshrew-alola',
    'sandshrew',
    'sandslash-alola',
    'sandslash',
    'sandygast',
    'sandy-shocks',
    'sawk',
    'sawsbuck',
    'scatterbug',
    'sceptile-mega',
    'sceptile',
    'scizor-mega',
    'scizor',
    'scolipede',
    'scorbunny',
    'scovillain',
    'scrafty',
    'scraggy',
    'scream-tail',
    'scyther',
    'seadra',
    'seaking',
    'sealeo',
    'seedot',
    'seel',
    'seismitoad',
    'sentret',
    'serperior',
    'servine',
    'seviper',
    'sewaddle',
    'sharpedo-mega',
    'sharpedo',
    'shaymin',
    'shaymin-sky',
    'shedinja',
    'shelgon',
    'shellder',
    'shellos',
    'shelmet',
    'shieldon',
    'shiftry',
    'shiinotic',
    'shinx',
    'shroodle',
    'shroomish',
    'shuckle',
    'shuppet',
    'sigilyph',
    'silcoon',
    'silicobra',
    'silvally',
    'simipour',
    'simisage',
    'simisear',
    'sinistcha',
    'sinistea',
    'sirfetchd',
    'sizzlipede',
    'skarmory',
    'skeledirge',
    'skiddo',
    'skiploom',
    'skitty',
    'skorupi',
    'skrelp',
    'skuntank',
    'skwovet',
    'slaking',
    'slakoth',
    'sliggoo-hisui',
    'sliggoo',
    'slither-wing',
    'slowbro-galar',
    'slowbro-mega',
    'slowbro',
    'slowking-galar',
    'slowking',
    'slowpoke-galar',
    'slowpoke',
    'slugma',
    'slurpuff',
    'smeargle',
    'smoliv',
    'smoochum',
    'sneasel-hisui',
    'sneasel',
    'sneasler',
    'snivy',
    'snom',
    'snorlax-gmax',
    'snorlax',
    'snorunt',
    'snover',
    'snubbull',
    'sobble',
    'solgaleo',
    'solosis',
    'solrock',
    'spearow',
    'spectrier',
    'spewpa',
    'spheal',
    'spidops',
    'spinarak',
    'spinda',
    'spiritomb',
    'spoink',
    'sprigatito',
    'spritzee',
    'squawkabilly-green',
    'squirtle',
    'stakataka',
    'stantler',
    'staraptor',
    'staravia',
    'starly',
    'starmie',
    'staryu',
    'steelix-mega',
    'steelix',
    'steenee',
    'stonjourner',
    'stoutland',
    'stufful',
    'stunfisk-galar',
    'stunfisk',
    'stunky',
    'sudowoodo',
    'suicune',
    'sunflora',
    'sunkern',
    'surskit',
    'swablu',
    'swadloon',
    'swalot',
    'swampert-mega',
    'swampert',
    'swanna',
    'swellow',
    'swinub',
    'swirlix',
    'swoobat',
    'sylveon',
    'tadbulb',
    'taillow',
    'talonflame',
    'tandemaus',
    'tangela',
    'tangrowth',
    'tapu-bulu',
    'tapu-fini',
    'tapu-koko',
    'tapu-lele',
    'tarountula',
    'tatsugiri-droopy',
    'tatsugiri',
    'tatsugiri-stretchy',
    'tauros',
    'teddiursa',
    'tentacool',
    'tentacruel',
    'tepig',
    'terapagos-terastal',
    'terrakion',
    'thievul',
    'throh',
    'thundurus',
    'thundurus-therian',
    'thwackey',
    'timburr',
    'ting-lu',
    'tinkatink',
    'tinkaton',
    'tinkatuff',
    'tirtouga',
    'toedscool',
    'toedscruel',
    'togedemaru',
    'togekiss',
    'togepi',
    'togetic',
    'torchic',
    'torkoal',
    'tornadus',
    'tornadus-therian',
    'torracat',
    'torterra',
    'totodile',
    'toucannon',
    'toxapex',
    'toxel',
    'toxicroak',
    'toxtricity-gmax',
    'toxtricity-low-key',
    'toxtricity',
    'tranquill',
    'trapinch',
    'treecko',
    'trevenant',
    'tropius',
    'trubbish',
    'trumbeak',
    'tsareena',
    'turtonator',
    'turtwig',
    'tympole',
    'tynamo',
    'type-null',
    'typhlosion-hisui',
    'typhlosion',
    'tyranitar-mega',
    'tyranitar',
    'tyrantrum',
    'tyrogue',
    'tyrunt',
    'umbreon',
    'unfezant',
    'unown',
    'ursaluna',
    'ursaluna-bloodmoon',
    'ursaring',
    'urshifu-gmax',
    'urshifu',
    'urshifu-rapid-strike-gmax',
    'uxie',
    'vanillish',
    'vanillite',
    'vanilluxe',
    'vaporeon',
    'varoom',
    'veluza',
    'venipede',
    'venomoth',
    'venonat',
    'venusaur-gmax',
    'venusaur-mega',
    'venusaur',
    'vespiquen',
    'vibrava',
    'victini',
    'victreebel',
    'vigoroth',
    'vikavolt',
    'vileplume',
    'virizion',
    'vivillon',
    'volbeat',
    'volcanion',
    'volcarona',
    'voltorb-hisui',
    'voltorb',
    'vullaby',
    'vulpix-alola',
    'vulpix',
    'wailmer',
    'wailord',
    'walking-wake',
    'walrein',
    'wartortle',
    'watchog',
    'wattrel',
    'weavile',
    'weedle',
    'weepinbell',
    'weezing-galar',
    'weezing',
    'whimsicott',
    'whirlipede',
    'whiscash',
    'whismur',
    'wigglytuff',
    'wiglett',
    'wimpod',
    'wingull',
    'wishiwashi',
    'wishiwashi-school',
    'wobbuffet',
    'wo-chien',
    'woobat',
    'wooloo',
    'wooper-paldea',
    'wooper',
    'wormadam',
    'wormadam-sandy',
    'wormadam-trash',
    'wugtrio',
    'wurmple',
    'wynaut',
    'wyrdeer',
    'xatu',
    'xerneas-active',
    'xerneas',
    'xurkitree',
    'yamask-galar',
    'yamask',
    'yamper',
    'yanma',
    'yanmega',
    'yungoos',
    'yveltal',
    'zacian-crowned',
    'zacian',
    'zamazenta-crowned',
    'zamazenta',
    'zangoose',
    'zapdos-galar',
    'zapdos',
    'zarude',
    'zebstrika',
    'zekrom',
    'zeraora',
    'zigzagoon-galar',
    'zigzagoon',
    'zoroark-hisui',
    'zoroark',
    'zorua-hisui',
    'zorua',
    'zubat',
    'zweilous',
    'zygarde-10',
    'zygarde-complete',
    'zygarde',
]

pokemon_as_decks = [{
    'name': p.replace('-', ' ').title(),
    'id': p,
    'icons': [p],
} for p in pokemon]



from . import utils

def fetch_row_info(row):
    """ given a row, fetch information """
    cells = row.findAll('td')
    placement = int(cells[0].get_text())
    name = cells[1].get_text().strip()
    country = cells[2].find('img')['alt']
    decklist_cell = cells[-1].find('a')
    decklist_url = decklist_cell['href'] if decklist_cell else None
    archetype = cells[-2].find('span')['data-tooltip']
    archetype_icons = [i['alt'] for i in cells[-2].findAll('img')]
    return placement, archetype, archetype_icons, decklist_url, name, country


def fetch_events():
    base_url = 'https://limitlesstcg.com/tournaments?show=300&time=all'
    page_html = utils.get_html(base_url)
    rows = utils.extract_table_rows(page_html, 'completed-tournaments')
    events = []
    for row in rows:
        name = row.findAll('a')[0]
        url = name['href'].split('/')[-1]
        events.append((name.text, url))
    return events


def fetch_decklists(tour_id):
    # TODO add in Juniors and Seniors here
    base_url = 'https://limitlesstcg.com'
    tour_url = f'{base_url}/tournaments/{tour_id}'

    html = utils.get_html(tour_url)
    rows = utils.extract_table_rows(html, 'data-table')
    decklists = [fetch_row_info(row) for row in rows]
    return decklists


def fetch_decklist(url):
    """ fetch a decklist from a given url """
    html = utils.get_html(url)
    soup_cards = html.findAll('div', {'class': 'decklist-card'})
    cards = []
    for soup_card in soup_cards:
        cards.append(
            {
                'number': soup_card['data-number'],
                'set': soup_card['data-set'],
                'count': int(soup_card.find('span', {'class': 'card-count'}).get_text()),
                'name': soup_card.find('span', {'class': 'card-name'}).get_text()
            }
        )
    return cards


def prompt_to_get_limitless_tour_id():
    print('* Finding Tournaments on Limitless')
    events = fetch_events()
    ids = {}
    for i, row in enumerate(events[:10]):
        ids[i] = row[1]
        print(f'[{i}] {row[0]}')
    selected_id = input('Select tournament: ')
    print('')
    return ids[int(selected_id)]


def fetch_decks():
    url = 'https://limitlesstcg.com/decks?variants=true'
    html = utils.get_html(url)
    rows = utils.extract_table_rows(html, 'data-table')
    # TODO extract decks
    return


if __name__ == '__main__':
    out = prompt_to_get_limitless_tour_id()
    print(out)
    decks = fetch_decklists(out)
    print(decks[0])



from dash import html, dcc
import dash_bootstrap_components as dbc
import inspect
import math
from typing import Callable

from th_helpers.components import deck_label
from th_helpers.utils import colors

win_rate_calc_comp = '''
$$
\% = \\frac{wins + \\frac{ties}{3}}{total}
$$
'''
SIGNIFICANT_MAPPING = {
    'some': '*',
    'all': '**',
    'favored': '*',
    'unfavored': '*'
}


def determine_win_rate(match):
    if 'Win' in match:
        wins = match.get('Win', 0)
        loss = match.get('Loss', 0)
        ties = match.get('Tie', 0)
    elif 'wins' in match:
        wins = match.get('wins', 0)
        loss = match.get('losses', 0)
        ties = match.get('ties', 0)
    elif 'w' in match:
        wins = match.get('w', 0)
        loss = match.get('l', 0)
        ties = match.get('t', 0)
    else:
        wins = 0
        loss = 0
        ties = 0
    total = wins + loss + ties
    if total == 0: return 0
    return round((wins + ties/3) / total * 100, 1)


def create_record_string(match):
    if 'Win' in match and 'Loss' in match:
        tied = f'-{match["Tie"]}' if match.get('Tie', 0) > 0 in match else ''
        record_string = f'{match["Win"]}-{match["Loss"]}{tied}'
        return record_string
    if 'wins' in match and 'losses' in match:
        tied = f'-{match["ties"]}' if match.get('ties', 0) > 0 in match else ''
        record_string = f'{match["wins"]}-{match["losses"]}{tied}'
        return record_string
    if 'w' in match or 'l' in match or 't' in match:
        tied = f'-{match["t"]}' if match.get('t', 0) > 0 else ''
        record_string = f'{match.get("w", 0)}-{match.get("l", 0)}{tied}'
        return record_string
    return None


def _call_label_func(func, deck, hide_text=False):
    if func is None:
        return None
    kwargs = {}
    kwargs['hide_text'] = hide_text
    if isinstance(func, Callable):
        return func(deck, **kwargs)
    return None


def create_popover_inside(color='', record='', wr='', decks=None, match=None, player=None, against=None, label_func=None):
    vs_item = html.Div([
        html.Span(
            _call_label_func(label_func, decks.get(match[player]), hide_text=True)
        ),
        html.Span('vs.', className='mx-2'),
        html.Span(
            _call_label_func(label_func, decks.get(match[against]), hide_text=True)
        ),
    ], className='d-flex align-items-center justify-content-around')
    return html.Div([
        vs_item,
        html.Div(f'{wr}%'),
        html.Div(record)
    ], className='text-black text-center p-2 rounded', style={'backgroundColor': color})


def create_record_display(match, className=None):
    wr = match['win_rate']
    record = create_record_string(match)
    color = colors.win_rate_color_bar[math.floor(wr)][1]
    return html.Td([
        f'{wr}%', html.Div(record)
    ], className=f'text-center text-black {className}', style={'backgroundColor': color})


def create_matchup_tile(match, decks, player, against, label_func=None):
    if match is None or match['win_rate'] is None or math.isnan(match['win_rate']):
        return html.Td('-', className='text-center align-middle')
    id = match[player] + match[against]
    wr = match['win_rate']
    record = create_record_string(match)
    color = colors.win_rate_color_bar[math.floor(wr)][1]
    return html.Td([
        html.Div([f'{wr}%', html.Div(record)], id=id, className='text-center'),
        dbc.Popover(
            create_popover_inside(color=color, record=record, wr=wr,
                                  decks=decks, match=match, player=player,
                                  against=against, label_func=label_func),
            target=id,
            trigger='hover',
            placement='bottom'
        ),
    ], style={'backgroundColor': color, 'width': '112px'}, className='text-center text-black align-middle')

def create_matchup_table_row(deck, data, decks, player, against, label_func=None):
    matches = [create_matchup_tile(match, decks, player, against, label_func=label_func) for match in data]
    row = html.Tr([html.Td(
        _call_label_func(label_func, decks[deck]),
        className='text-nowrap align-middle'
    )] + matches)
    return row

def create_matchup_tile_full(match, decks, player, against, label_func=None):
    if match is None or match['win_rate'] is None or math.isnan(match['win_rate']):
        return html.Span(className='d-none')
    id = match[player] + match[against]
    wr = match['win_rate']
    record = create_record_string(match)
    color = colors.win_rate_color_bar[math.floor(wr)][1]
    vs_item = html.Div([
        html.Span('vs.', className='me-1'),
        html.Span(
            _call_label_func(label_func, decks.get(match[against]), hide_text=True)
        ),
    ], className='d-flex align-items-center')
    significant = SIGNIFICANT_MAPPING[match['significant']] if match.get('significant', None) else ''
    return dbc.Card(
        dbc.CardBody([
            vs_item,
            html.Div(f'{match["win_rate"]}%{significant}'),
            html.Div(record)
        ], class_name='text-black text-center p-1'),
        style={'backgroundColor': color},
        className='w-auto',
        id=id
    )    

def create_matchup_tile_row(deck, data, decks, player, against, label_func=None):
    row = html.Div([
        html.H5(_call_label_func(label_func, decks[deck])),
        dbc.Row([create_matchup_tile_full(match, decks, player, against, label_func=label_func) for match in data], class_name='g-1')
    ], className='mb-2')
    return row

def create_matchup_spread(data, decks, player='deck1', against='deck2', small_view=False, label_func=None):
    # Extract unique decks from player and sort them alphabetically
    player_unique_decks = list(set(matchup[player] for matchup in data))
    if len(player_unique_decks) == 0:
        return 'No matchup information found.'
    if 'Plays:' in player_unique_decks[0]:
        player_unique_decks = sorted(player_unique_decks, key=lambda x: int(x.split(':')[1].strip()))
    else:
        player_unique_decks = sorted(player_unique_decks)
    against_unique_decks = sorted(set(matchup[against] for matchup in data))

    rows = []
    small_rows = []
    # Organize the data
    for deck in player_unique_decks:
        if deck not in decks:
            icons = ['substitute'] if 'Plays:' not in deck else []
            decks[deck] = {'id': deck, 'name': deck.title(), 'icons': icons}
        matchups = sorted(
            (matchup for matchup in data if matchup[player] == deck),
            key=lambda x: x[against]
        )
        duplicates = [index for index, d in enumerate(matchups) if d[player] == d[against]]
        if len(duplicates) > 1:
            matchups.pop(duplicates[0])

        ordered_matchups = [None for _ in range(len(against_unique_decks))]
        for m in matchups:
            ordered_matchups[against_unique_decks.index(m[against])] = m
        rows.append(create_matchup_table_row(deck, ordered_matchups, decks, player, against, label_func=label_func))
        small_rows.append(create_matchup_tile_row(deck, ordered_matchups, decks, player, against, label_func=label_func))

    header_labels = [
        html.Div(
            _call_label_func(label_func, decks.get(deck, deck_label.create_default_deck(deck)), hide_text=True),
            className='d-flex justify-content-center')
        for deck in against_unique_decks
    ]
    headers = html.Thead(html.Tr([
        html.Th(deck) for deck in [win_rate_calc_comp] + header_labels
    ]), className='sticky-top')
    table = dbc.Table([
        headers,
        html.Tbody(rows)
    ], className='d-none' if small_view else 'd-none d-xl-block')

    small_view = html.Div([
        # win_rate_calc_comp,
        html.Div(small_rows)
    ], className='' if small_view else 'd-xl-none')
    return html.Div([table, small_view])


def create_example(win_rate_mathjax):
    example = html.Div([
        html.Div([
            'Other vs Other',
            create_popover_inside('#fff', 'Wins-Losses-Ties', 'rate', decks={}, match={None: None}),
            'Color based on rate',
            html.Div(style={
                'background': f'linear-gradient(to right, {", ".join(colors.red_to_white_to_blue)})',
                'height': '20px',
                'width': '100%',
                'border': '1px solid #ccc'
            }),
            html.Div([html.Span(c) for c in [0, 100]], className='d-flex justify-content-between')
        ], className='text-center'),
        html.Div([
            'Weighted success rate',
            dcc.Markdown(
                win_rate_mathjax,
                mathjax=True, className='win-rate-calc'
            )
        ]),
    ])
    return example



class TIME:
    MINUTE = 60
    HOUR = MINUTE * 60
    HALF_DAY = HOUR * 12
    DAY = HOUR * 24


class DASH:
    '''Constants used for referencing different component
    attributes - primarly for callbacks.
    This list includes attributes for dash.html, dash.dcc,
    and dash-bootstrap-components.
    '''
    ACTIVE = 'active'
    BASE64_IMAGE_STR = 'data:image/png;base64,{}'
    BLANK = '_blank'
    CHILDREN = 'children'
    CLASSNAME = 'className'
    CLASS_NAME = 'class_name'
    CLIENTSIDE = 'clientside'  # script namespace
    COLOR = 'color'
    DATA = 'data'
    DISABLED = 'disabled'
    END_DATE = 'end_date'
    HEADER = 'header'
    HREF = 'href'
    ICON = 'icon'
    ID = 'id'
    INDEX = 'index'
    INITIAL_DUPLICATE = 'initial_duplicate'
    INVALID = 'invalid'
    IS_OPEN = 'is_open'
    LABEL = 'label'
    LAST = 'last'
    MAXLENGTH = 'maxlength'
    N_CLICKS = 'n_clicks'
    NUMBER = 'number'
    OPTIONS = 'options'
    ORDER = 'order'
    PAGES_LOCATION = '_pages_location'
    PATHNAME = 'pathname'
    SIZE = 'size'
    SRC = 'src'
    START_DATE = 'start_date'
    SUBMIT_N_CLICKS = 'submit_n_clicks'
    SUBTYPE = 'subtype'
    TYPE = 'type'
    VALUE = 'value'


class BOOTSTRAP:
    '''Constants specific to Bootstrap
    These are mostly used for coloring or positioning.
    '''
    class ALIGNMENT:
        BOTTOM_END = 'bottom-end'
        CENTER = 'center'
        END = 'end'

    class COLOR:
        DANGER = 'danger'
        DARK = 'dark'
        INFO = 'info'
        LIGHT = 'light'
        PRIMARY = 'primary'
        SECONDARY = 'secondary'
        SUCCESS = 'success'
        TRANSPARENT = 'transparent'
        WARNING = 'warning'


class STYLE:
    '''Constants used when defining inline styles
    '''
    BACKGROUND = 'background'
    HEIGHT = 'height'
    MAXHEIGHT = 'maxHeight'
    MAXWIDTH = 'maxWidth'
    WIDTH = 'width'



from dash import html, clientside_callback, Output, Input, State, MATCH
import dash_bootstrap_components as dbc
import uuid

HTML2CANVAS_URL = {'src': 'https://cdnjs.cloudflare.com/ajax/libs/html2canvas/1.4.1/html2canvas.min.js'}


class DownloadImageAIO(html.Div):

    class ids:
        button = lambda aio_id: {
            'component': 'DownloadImageAIO',
            'subcomponent': 'button',
            'aio_id': aio_id
        }
        dom_id = lambda aio_id: {
            'component': 'DownloadImageAIO',
            'subcomponent': 'dom_id',
            'aio_id': aio_id
        }
        dummy = lambda aio_id: {
            'component': 'DownloadImageAIO',
            'subcomponent': 'dummy',
            'aio_id': aio_id
        }
    ids = ids

    def __init__(
        self,
        dom_id=None,
        aio_id=None,
        className='',
        button_class_name='',
        text='Download'
    ):
        if aio_id is None:
            aio_id = str(uuid.uuid4())
        
        button = [
            dbc.Button([
                    html.I(className='fas fa-download', title='Download image (png)'),
                    html.Span(text, className='ms-1 d-sm-inline-block d-none') if text else ''
                ],
                id=self.ids.button(aio_id),
                n_clicks=0,
                class_name=button_class_name
            ),
            dbc.Input(id=self.ids.dom_id(aio_id), value=dom_id, class_name='d-none'),
            html.Div(id=self.ids.dummy(aio_id))
        ]
        super().__init__(button, className=className)

    clientside_callback(
        '''
        function (clicks, id) {
          const today = new Date();
          const dateString = today.toISOString().substring(0, 10);
          fileName = `trainerhill-${id}-${dateString}.png`;
          if (clicks > 0) {
            html2canvas(document.getElementById(id), { useCORS: true, backgroundColor: '#ffffff' }).then(function (canvas) {
              var anchorTag = document.createElement('a');
              anchorTag.download = fileName;
              anchorTag.href = canvas.toDataURL('image/png');
              anchorTag.target = '_blank';
              document.body.appendChild(anchorTag);
              anchorTag.click();
              document.body.removeChild(anchorTag);
            })
        }
        ''',
        Output(ids.dummy(MATCH), 'className'),
        Input(ids.button(MATCH), 'n_clicks'),
        State(ids.dom_id(MATCH), 'value'),
    )



# package imports
from bs4 import BeautifulSoup
from contextlib import closing
from requests import get


def get_html(url):
    """ scrapes a webpage and returns the beautified soup """
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/118.0 Safari/537.36"
        ),
        "Accept-Language": "en-US,en;q=0.9",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Connection": "keep-alive",
    }
    with closing(get(url, headers=headers, stream=True, timeout=30)) as resp:
        resp.raise_for_status()
        return BeautifulSoup(resp.content, 'html.parser')


def extract_table_rows(html, class_name):
    """ extract the table from beautiful soup data given class name """
    table = html.find("table", {"class": class_name})
    if not table:
        return []  # or raise a custom exception
    rows = list(table.find_all("tr"))
    return rows[1:] if len(rows) > 1 else rows



from dash import html


pokemon_url = 'https://raw.githubusercontent.com/bradley-erickson/pokesprite/master/pokemon/regular'

pokemon_mapping = {
    'ogerpon': 'ogerpon-teal-mask',
    'squawkabilly': 'squawkabilly-green',
    'regieleki-a': 'regieleki',
    'ogerpon-cornerstone': 'ogerpon-cornerstone-mask',
    'ogerpon-wellspring': 'ogerpon-wellspring-mask',
    'ogerpon-hearthflame': 'ogerpon-hearthflame-mask',
    'terapagos': 'terapagos-terastal'
}


# TODO add game
def format_label(deck, hide_text=False, hide_text_small=False):
    if deck is None:
        return ''
    children = [
        html.Img(
            src=i,
            style={'maxHeight': '35px'}
        ) for i in deck.get('icons', [])
    ]
    name = deck.get('name')
    children.append(html.Span(name, className='d-none' if hide_text else 'd-none d-md-inline-block' if hide_text_small else 'ms-1'))
    return html.Div(
        children, title=name,
        className='d-flex flex-row align-items-center'
    )


def get_pokemon_icon(pokemon):
    if pokemon.startswith('https'):
        return pokemon
    if not pokemon:
        return ''
    if pokemon == 'substitute':
        source = '/assets/substitute.png'
    else:
        mon = pokemon_mapping[pokemon] if pokemon in pokemon_mapping else pokemon
        source = f'{pokemon_url}/{mon}.png'
    return source


def create_default_deck(id):
    id_fix = 'other' if id is None else id
    return {'id': id_fix, 'name': id_fix.title(), 'icons': ['substitute']}



'''
This file was automatically generated by
`th-scripting/palette_extractor.py`.
'''
ICON_COLOR_MAPPING = {
	"abomasnow-mega": "#4F6057",
	"abomasnow": "#F7F6F6",
	"abra": "#EED845",
	"absol-mega": "#F3F3F3",
	"absol": "#575C6A",
	"accelgor": "#575057",
	"aegislash-blade": "#60594F",
	"aegislash": "#5C554A",
	"aerodactyl-mega": "#3E3A3E",
	"aerodactyl": "#83787B",
	"aggron-mega": "#505050",
	"aggron": "#4E4F4F",
	"aipom": "#B097B7",
	"alakazam-mega": "#746459",
	"alakazam": "#7C6A4E",
	"alcremie-caramel-swirl-berry": "#E9D2B1",
	"alcremie-caramel-swirl-clover": "#CCAD7F",
	"alcremie-caramel-swirl-flower": "#A76D44",
	"alcremie-caramel-swirl-love": "#DBA485",
	"alcremie-caramel-swirl-plain": "#EDD4B1",
	"alcremie-caramel-swirl-ribbon": "#BB9485",
	"alcremie-caramel-swirl-star": "#9E7C54",
	"alcremie-caramel-swirl-strawberry": "#C18D6C",
	"alcremie-gmax": "#EFE9E0",
	"alcremie-lemon-cream-berry": "#F3EEAF",
	"alcremie-lemon-cream-clover": "#CABF5E",
	"alcremie-lemon-cream-flower": "#DEC863",
	"alcremie-lemon-cream-love": "#DCAF69",
	"alcremie-lemon-cream-plain": "#F6F0AF",
	"alcremie-lemon-cream-ribbon": "#F4ECB7",
	"alcremie-lemon-cream-star": "#DBCA48",
	"alcremie-lemon-cream-strawberry": "#DAC565",
	"alcremie-matcha-cream-berry": "#CFE497",
	"alcremie-matcha-cream-clover": "#97AE63",
	"alcremie-matcha-cream-flower": "#CDE08F",
	"alcremie-matcha-cream-love": "#BAA873",
	"alcremie-matcha-cream-plain": "#D6EA9C",
	"alcremie-matcha-cream-ribbon": "#9FA27A",
	"alcremie-matcha-cream-star": "#B5C053",
	"alcremie-matcha-cream-strawberry": "#A3A867",
	"alcremie-mint-cream-berry": "#CBEEF0",
	"alcremie-mint-cream-clover": "#C9EDEF",
	"alcremie-mint-cream-flower": "#C2E7E9",
	"alcremie-mint-cream-love": "#C3E6E9",
	"alcremie-mint-cream-plain": "#CBEEF1",
	"alcremie-mint-cream-ribbon": "#CAECEF",
	"alcremie-mint-cream-star": "#C2E7E8",
	"alcremie-mint-cream-strawberry": "#C9ECEE",
	"alcremie": "#F9E5E0",
	"alcremie-rainbow-swirl-berry": "#EADAB1",
	"alcremie-rainbow-swirl-clover": "#AEA56A",
	"alcremie-rainbow-swirl-flower": "#ECD9AB",
	"alcremie-rainbow-swirl-love": "#D69A6C",
	"alcremie-rainbow-swirl-plain": "#EFE2BB",
	"alcremie-rainbow-swirl-ribbon": "#EDD9B2",
	"alcremie-rainbow-swirl-star": "#EADBB5",
	"alcremie-rainbow-swirl-strawberry": "#BEA370",
	"alcremie-ruby-cream-berry": "#EEB2D2",
	"alcremie-ruby-cream-clover": "#F2B7D5",
	"alcremie-ruby-cream-flower": "#F1B2D0",
	"alcremie-ruby-cream-love": "#F3B7D3",
	"alcremie-ruby-cream-plain": "#F2B9D8",
	"alcremie-ruby-cream-ribbon": "#BF76A0",
	"alcremie-ruby-cream-star": "#F0B3CF",
	"alcremie-ruby-cream-strawberry": "#F2B5D4",
	"alcremie-ruby-swirl-berry": "#F0C0AB",
	"alcremie-ruby-swirl-clover": "#F6CBAF",
	"alcremie-ruby-swirl-flower": "#F4C5A9",
	"alcremie-ruby-swirl-love": "#D87F83",
	"alcremie-ruby-swirl-plain": "#F7D2B1",
	"alcremie-ruby-swirl-ribbon": "#C37F8F",
	"alcremie-ruby-swirl-star": "#F3C2A8",
	"alcremie-ruby-swirl-strawberry": "#F6C9AE",
	"alcremie-salted-cream-berry": "#E4F0E9",
	"alcremie-salted-cream-clover": "#E5F0E9",
	"alcremie-salted-cream-flower": "#E1ECE4",
	"alcremie-salted-cream-love": "#E4ECE6",
	"alcremie-salted-cream-plain": "#E8F2EB",
	"alcremie-salted-cream-ribbon": "#E4EFE9",
	"alcremie-salted-cream-star": "#E1EDE4",
	"alcremie-salted-cream-strawberry": "#E6F1EA",
	"alcremie-vanilla-cream-berry": "#F7E5DF",
	"alcremie-vanilla-cream-clover": "#F8E5DF",
	"alcremie-vanilla-cream-flower": "#F9E7E0",
	"alcremie-vanilla-cream-love": "#F8E3DD",
	"alcremie-vanilla-cream-plain": "#F9E8E2",
	"alcremie-vanilla-cream-ribbon": "#F8E4DF",
	"alcremie-vanilla-cream-star": "#F6E1D9",
	"alcremie-vanilla-cream-strawberry": "#F9E5DF",
	"alomomola": "#F9B7D6",
	"altaria-mega": "#DBE3F3",
	"altaria": "#A2BECA",
	"amaura": "#8AC2EB",
	"ambipom": "#584357",
	"amoonguss": "#CD7E88",
	"ampharos-mega": "#665545",
	"ampharos": "#E8CB41",
	"annihilape": "#BBB9BD",
	"anorith": "#4A4844",
	"appletun-gmax": "#6F4336",
	"appletun": "#E4B268",
	"applin": "#D24E56",
	"araquanid": "#525F62",
	"arbok": "#B293CA",
	"arboliva": "#465825",
	"arcanine-hisui-noble": "#DBC430",
	"arcanine-hisui": "#746969",
	"arcanine": "#CB9E70",
	"arceus-bug": "#606941",
	"arceus-dark": "#565858",
	"arceus-dragon": "#53545B",
	"arceus-electric": "#E6E5E1",
	"arceus-fairy": "#F5ECEF",
	"arceus-fighting": "#211817",
	"arceus-fire": "#EAE9E7",
	"arceus-flying": "#989FB0",
	"arceus-ghost": "#584D5B",
	"arceus-grass": "#536F48",
	"arceus-ground": "#EAE8E7",
	"arceus-ice": "#95AFBA",
	"arceus": "#5D5F4B",
	"arceus-poison": "#624E59",
	"arceus-psychic": "#615353",
	"arceus-rock": "#6D683E",
	"arceus-steel": "#595A58",
	"arceus-unknown": "#4B5956",
	"arceus-water": "#3E5262",
	"archaludon": "#121629",
	"archen": "#AD5F46",
	"archeops": "#207FA1",
	"arctibax": "#748C9F",
	"arctovish": "#ACD0EE",
	"arctozolt": "#5B8390",
	"ariados": "#5F4646",
	"armaldo": "#555558",
	"armarouge": "#1A130F",
	"aromatisse": "#AF72AE",
	"aron": "#C8D0D0",
	"arrokuda": "#9D8873",
	"articuno-galar": "#C4AECC",
	"articuno": "#83C3F0",
	"audino-mega": "#F2EBDB",
	"audino": "#E9A09F",
	"aurorus": "#7396C3",
	"avalugg-hisui-noble": "#D3C937",
	"avalugg-hisui": "#74615B",
	"avalugg": "#E4ECF4",
	"axew": "#51763A",
	"azelf": "#569DC8",
	"azumarill": "#63B7ED",
	"azurill": "#82B3EC",
	"bagon": "#555C68",
	"baltoy": "#B37A58",
	"banette-mega": "#7B7274",
	"banette": "#878179",
	"barbaracle": "#615343",
	"barboach": "#62778F",
	"barraskewda": "#A37D64",
	"basculegion": "#4C7355",
	"basculin-blue-striped": "#5F7E5B",
	"basculin": "#7B875F",
	"basculin-white-striped": "#566B52",
	"bastiodon": "#828078",
	"baxcalibur": "#495568",
	"bayleef": "#E3DE67",
	"beartic": "#E1F2FA",
	"beautifly": "#6A6A6A",
	"beedrill-mega": "#5E5A52",
	"beedrill": "#6D6157",
	"beheeyem": "#988165",
	"beldum": "#6087CF",
	"bellibolt": "#2F916B",
	"bellossom": "#6A8A54",
	"bellsprout": "#EDE457",
	"bergmite": "#A9CDE7",
	"bewear": "#595856",
	"bibarel": "#615243",
	"bidoof": "#BE9661",
	"binacle": "#A67959",
	"bisharp": "#775246",
	"blacephalon": "#F1E0EE",
	"blastoise-gmax": "#545552",
	"blastoise-mega": "#595147",
	"blastoise": "#6E7474",
	"blaziken-mega": "#BE6C4B",
	"blaziken": "#C1734D",
	"blipbug": "#71708D",
	"blissey": "#F3C9D6",
	"blitzle": "#5E605F",
	"boldore": "#434B74",
	"boltund": "#63674D",
	"bombirdier": "#D8D7D7",
	"bonsly": "#719157",
	"bouffalant": "#5E5341",
	"bounsweet": "#C95B89",
	"braixen": "#1E1612",
	"brambleghast": "#6D4E35",
	"bramblin": "#745A40",
	"braviary-hisui": "#31313A",
	"braviary": "#50557B",
	"breloom": "#5C5640",
	"brionne": "#5E9CCE",
	"bronzong": "#5B8181",
	"bronzor": "#538996",
	"brute-bonnet": "#88443F",
	"bruxish": "#BA71B0",
	"budew": "#8AB851",
	"buizel": "#DD8C36",
	"bulbasaur": "#4D7747",
	"buneary": "#B4855D",
	"bunnelby": "#81716A",
	"burmy": "#54604F",
	"burmy-sandy": "#E9DF74",
	"burmy-trash": "#ED928C",
	"butterfree-gmax": "#8CD4B7",
	"butterfree": "#555363",
	"buzzwole": "#63423E",
	"cacnea": "#629252",
	"cacturne": "#4A7048",
	"calyrex-ice-rider": "#365C4F",
	"calyrex": "#335D4B",
	"calyrex-shadow-rider": "#38564D",
	"camerupt-mega": "#C96856",
	"camerupt": "#E17E5C",
	"capsakid": "#759767",
	"carbink": "#7D89A0",
	"carkol": "#534E4A",
	"carnivine": "#6A8549",
	"carracosta": "#484E5C",
	"carvanha": "#C18D47",
	"cascoon": "#DABEEF",
	"castform": "#CDCDCC",
	"castform-rainy": "#78A2F4",
	"castform-snowy": "#446C60",
	"castform-sunny": "#B3613E",
	"caterpie": "#655635",
	"celebi": "#8AB06E",
	"celesteela": "#649D84",
	"centiskorch-gmax": "#724524",
	"centiskorch": "#653B2C",
	"ceruledge": "#1E1F2D",
	"cetitan": "#8C868A",
	"cetoddle": "#E3DBE2",
	"chandelure": "#474649",
	"chansey": "#D78A86",
	"charcadet": "#503D39",
	"charizard-gmax": "#F4A137",
	"charizard-mega-x": "#57606C",
	"charizard-mega-y": "#C76051",
	"charizard": "#C1675C",
	"charjabug": "#91AC67",
	"charmander": "#E37D37",
	"charmeleon": "#DE5E46",
	"chatot": "#464443",
	"cherrim": "#9571CA",
	"cherrim-sunshine": "#E28898",
	"cherubi": "#DB6A8F",
	"chesnaught": "#9A9966",
	"chespin": "#5D5E39",
	"chewtle": "#B6D5CD",
	"chien-pao": "#0A1019",
	"chikorita": "#7CBB56",
	"chimchar": "#D27F48",
	"chimecho": "#CFE1F5",
	"chinchou": "#A5B8DF",
	"chingling": "#1D1610",
	"chi-yu": "#663C2D",
	"cinccino": "#D4DADC",
	"cinderace-gmax": "#F8B12F",
	"cinderace": "#59494D",
	"clamperl": "#484752",
	"clauncher": "#63ADE3",
	"clawitzer": "#5397E1",
	"claydol": "#5E5352",
	"clefable": "#FACCCB",
	"clefairy": "#F8BEA8",
	"cleffa": "#F3B39E",
	"clobbopus": "#E7D6CA",
	"clodsire": "#61514E",
	"cloyster": "#6A6389",
	"coalossal-gmax": "#5F5955",
	"coalossal": "#554D4A",
	"cobalion": "#436768",
	"cofagrigus": "#414344",
	"combee": "#735E3A",
	"combusken": "#F7E784",
	"comfey": "#BCA05A",
	"conkeldurr": "#977669",
	"copperajah-gmax": "#42534F",
	"copperajah": "#2C4443",
	"corphish": "#D06C46",
	"corsola-galar": "#E9E9E8",
	"corsola": "#F29398",
	"corviknight-gmax": "#785A91",
	"corviknight": "#676197",
	"corvisquire": "#474950",
	"cosmoem": "#515155",
	"cosmog": "#4A4963",
	"cottonee": "#E0F2F3",
	"crabominable": "#E3DFDB",
	"crabrawler": "#505460",
	"cradily": "#585D4D",
	"cramorant-gorging": "#325C85",
	"cramorant-gulping": "#315C88",
	"cramorant": "#2F5D8C",
	"cranidos": "#565B62",
	"crawdaunt": "#C74F39",
	"cresselia": "#CE9FC0",
	"croagunk": "#818EDE",
	"crobat": "#6E6F97",
	"crocalor": "#7D3F35",
	"croconaw": "#6AB4D5",
	"crustle": "#745851",
	"cryogonal": "#6B8CCD",
	"cubchoo": "#7EBAEC",
	"cubone": "#ACA08E",
	"cufant": "#E18F63",
	"cursola": "#DED6EB",
	"cutiefly": "#D9C879",
	"cyclizar-mount": "#3E474D",
	"cyclizar": "#3E474D",
	"cyndaquil": "#19120E",
	"dachsbun": "#5E3326",
	"darkrai": "#423F3D",
	"darmanitan-galar": "#FAFBFB",
	"darmanitan-galar-zen": "#E9EBE8",
	"darmanitan": "#B45049",
	"darmanitan-zen": "#496062",
	"dartrix": "#60563E",
	"darumaka-galar": "#EBF2F4",
	"darumaka": "#C04949",
	"decidueye-hisui": "#D17657",
	"decidueye": "#585649",
	"dedenne": "#CE9324",
	"deerling-autumn": "#EBA843",
	"deerling": "#E5BAB2",
	"deerling-summer": "#83AA64",
	"deerling-winter": "#CBAD6A",
	"deino": "#81717E",
	"delcatty": "#AB79BD",
	"delibird": "#E2DFDB",
	"delphox": "#D74323",
	"deoxys-attack": "#544139",
	"deoxys-defense": "#E06F52",
	"deoxys": "#D46A51",
	"deoxys-speed": "#CB664F",
	"dewgong": "#F0F6FC",
	"dewott": "#567490",
	"dewpider": "#709E96",
	"dhelmise": "#4F5143",
	"dialga-origin": "#3D4F5E",
	"dialga": "#455262",
	"diancie-mega": "#A9888B",
	"diancie": "#BC94A3",
	"diggersby": "#7F694E",
	"diglett-alola": "#9B7A5E",
	"diglett": "#695945",
	"dipplin": "#4B3B2C",
	"ditto": "#967CCF",
	"dodrio": "#BA8C5C",
	"doduo": "#D0A976",
	"dolliv": "#478419",
	"dondozo": "#171E2F",
	"donphan": "#5C5F68",
	"dottler": "#E6C653",
	"doublade": "#734A53",
	"dracovish": "#405F5E",
	"dracozolt": "#B66F46",
	"dragalge": "#654338",
	"dragapult": "#596065",
	"dragonair": "#6790C9",
	"dragonite": "#EBB04E",
	"drakloak": "#546367",
	"drampa": "#98B39D",
	"drapion": "#8B6996",
	"dratini": "#819FC8",
	"drednaw-gmax": "#5F6E6C",
	"drednaw": "#655747",
	"dreepy": "#868981",
	"drifblim": "#ACA1D1",
	"drifloon": "#ADA1DB",
	"drilbur": "#877F8E",
	"drizzile": "#527085",
	"drowzee": "#E5CD3F",
	"druddigon": "#1C7BA6",
	"dubwool": "#4F4749",
	"ducklett": "#5A7AA7",
	"dudunsparce": "#DEDAA9",
	"dudunsparce-three-segment": "#DEDCAC",
	"dugtrio-alola": "#5D544C",
	"dugtrio": "#997565",
	"dunsparce": "#F1EBA0",
	"duosion": "#6EB576",
	"duraludon-gmax": "#BADADF",
	"duraludon": "#A6A29E",
	"durant": "#4F4D4B",
	"dusclops": "#817E71",
	"dusknoir": "#626052",
	"duskull": "#908E87",
	"dustox": "#B4A743",
	"dwebble": "#5F4D41",
	"eelektrik": "#4D5CA7",
	"eelektross": "#5B6ABB",
	"eevee-gmax": "#D2AF5E",
	"eevee": "#CCA558",
	"eevee-starter": "#D4A769",
	"eiscue-noice": "#CDE8F2",
	"eiscue": "#CCE2EA",
	"ekans": "#B3759D",
	"eldegoss": "#F9F7EE",
	"electabuzz": "#615755",
	"electivire": "#685E44",
	"electrike": "#82BF6F",
	"electrode-hisui-noble": "#F2E440",
	"electrode-hisui": "#E0C28F",
	"electrode": "#FAF9F9",
	"elekid": "#F2DD3F",
	"elgyem": "#A3E1DE",
	"emboar": "#7D6155",
	"emolga": "#716C66",
	"empoleon": "#31312E",
	"enamorus": "#62464A",
	"enamorus-therian": "#A1565F",
	"entei": "#A46956",
	"escavalier": "#B8AAA0",
	"espathra": "#C4AA8E",
	"espeon": "#535659",
	"espurr": "#A9A3BA",
	"eternatus-eternamax": "#433C43",
	"eternatus": "#403D57",
	"excadrill": "#8B716E",
	"exeggcute": "#DA9283",
	"exeggutor-alola": "#445046",
	"exeggutor": "#445147",
	"exploud": "#917F95",
	"falinks": "#342E2D",
	"farfetchd-galar": "#827766",
	"farfetchd": "#BAA661",
	"farigiraf": "#26211D",
	"fearow": "#A06D3A",
	"feebas": "#CAAA6F",
	"fennekin": "#B15D33",
	"feraligatr": "#53A1BE",
	"ferroseed": "#717F71",
	"ferrothorn": "#405548",
	"fezandipiti": "#604045",
	"fidough": "#C9B166",
	"finizen": "#50C2D7",
	"finneon": "#55626E",
	"flaaffy": "#C798A0",
	"flabebe-blue": "#1E231A",
	"flabebe-orange": "#E89E19",
	"flabebe": "#DB783D",
	"flabebe-white": "#DFDCDA",
	"flabebe-yellow": "#E5D02C",
	"flamigo": "#644248",
	"flapple-gmax": "#6F4336",
	"flapple": "#68412E",
	"flareon": "#F8EA9E",
	"fletchinder": "#655B55",
	"fletchling": "#4E4645",
	"flittle": "#F3E5A1",
	"floatzel": "#6C5B46",
	"floette-blue": "#1D2B23",
	"floette-eternal": "#23191B",
	"floette-orange": "#DF9417",
	"floette": "#23291E",
	"floette-white": "#E3E4E1",
	"floette-yellow": "#5D714B",
	"floragato": "#476439",
	"florges-blue": "#40756D",
	"florges-orange": "#487963",
	"florges": "#497168",
	"florges-white": "#427462",
	"florges-yellow": "#4E7C60",
	"flutter-mane": "#2E3D44",
	"flygon": "#5F6B4E",
	"fomantis": "#52574E",
	"foongus": "#A3685A",
	"forretress": "#DABDE5",
	"fraxure": "#74A46E",
	"frigibax": "#9A9569",
	"frillish": "#C3E4F8",
	"froakie": "#475857",
	"frogadier": "#3B4D68",
	"froslass": "#685861",
	"frosmoth": "#A5AAC0",
	"fuecoco": "#9B4C35",
	"furfrou-dandy": "#83B661",
	"furfrou-debutante": "#E5E1A4",
	"furfrou-diamond": "#262321",
	"furfrou-heart": "#48403F",
	"furfrou-kabuki": "#594540",
	"furfrou-la-reine": "#42504A",
	"furfrou-matron": "#564B58",
	"furfrou-pharaoh": "#444A5E",
	"furfrou": "#DCE3CB",
	"furfrou-star": "#445154",
	"furret": "#916F4E",
	"gabite": "#868ED1",
	"gallade-mega": "#575E55",
	"gallade": "#59856B",
	"galvantula": "#A19378",
	"garbodor-gmax": "#8F9272",
	"garbodor": "#67765A",
	"garchomp-mega": "#5777C0",
	"garchomp": "#5B649B",
	"gardevoir-mega": "#89A280",
	"gardevoir": "#82B79C",
	"garganacl": "#685F5D",
	"gastly": "#9276C2",
	"gastrodon-east": "#4E7651",
	"gastrodon": "#9D7C49",
	"genesect-burn": "#7D5896",
	"genesect-chill": "#7D5B9C",
	"genesect-douse": "#795897",
	"genesect-no-drive": "#AA80BC",
	"genesect": "#614874",
	"genesect-shock": "#7B599A",
	"gengar-gmax": "#943E87",
	"gengar-mega": "#7A60AA",
	"gengar": "#8C7EC5",
	"geodude-alola": "#4D4F52",
	"geodude": "#8A8365",
	"gholdengo": "#E5BA5A",
	"gible": "#6FA0B9",
	"gigalith": "#50516B",
	"gimmighoul": "#794F37",
	"gimmighoul-roaming": "#635850",
	"girafarig": "#785138",
	"giratina-origin": "#4B4643",
	"giratina": "#59534A",
	"glaceon": "#699CA5",
	"glalie-mega": "#454C4E",
	"glalie": "#535559",
	"glameow": "#9E969E",
	"glastrier": "#E4ECEF",
	"gligar": "#A16994",
	"glimmet": "#38456D",
	"glimmora": "#293D70",
	"gliscor": "#878BDA",
	"gloom": "#B16239",
	"gogoat": "#3F6947",
	"golbat": "#56617A",
	"goldeen": "#E3D9D4",
	"golduck": "#52687F",
	"golem-alola": "#404A57",
	"golem": "#897664",
	"golett": "#6B6651",
	"golisopod": "#585253",
	"golurk": "#6FACAB",
	"goodra-hisui": "#A8A3B4",
	"goodra": "#EADCF6",
	"goomy": "#C7ACE7",
	"gorebyss": "#F6A0C0",
	"gossifleur": "#BF5D32",
	"gothita": "#AE9AB2",
	"gothitelle": "#535453",
	"gothorita": "#5A5A5B",
	"gouging-fire": "#24272A",
	"gourgeist-large": "#D78A4E",
	"gourgeist": "#5F4940",
	"gourgeist-small": "#533C2C",
	"gourgeist-super": "#D8894F",
	"grafaiai": "#383A44",
	"granbull": "#D7A1D2",
	"grapploct": "#486DBE",
	"graveler-alola": "#45433F",
	"graveler": "#86806A",
	"great-tusk": "#494764",
	"greavard": "#131416",
	"greedent": "#90594D",
	"greninja-ash": "#445365",
	"greninja": "#3F5C91",
	"grimer-alola": "#576E4F",
	"grimer": "#BAA4D7",
	"grimmsnarl-gmax": "#675F91",
	"grimmsnarl": "#6D708F",
	"grookey": "#CAD07B",
	"grotle": "#4B5D3F",
	"groudon": "#604D4F",
	"groudon-primal": "#6B4953",
	"grovyle": "#3C6038",
	"growlithe-hisui": "#906059",
	"growlithe": "#E7BA77",
	"grubbin": "#635245",
	"grumpig": "#74686E",
	"gulpin": "#98CA81",
	"gumshoos": "#877449",
	"gurdurr": "#C67071",
	"guzzlord": "#474B4F",
	"gyarados-mega": "#615A59",
	"gyarados": "#536065",
	"hakamo-o": "#726C56",
	"happiny": "#F8D0C7",
	"hariyama": "#635358",
	"hatenna": "#CBD4E4",
	"hatterene-gmax": "#959CB3",
	"hatterene": "#D2D5E3",
	"hattrem": "#CFD8E7",
	"haunter": "#9477C4",
	"hawlucha": "#C17350",
	"haxorus": "#5E5242",
	"heatmor": "#6A584D",
	"heatran": "#5D4F45",
	"heliolisk": "#565246",
	"helioptile": "#8B845C",
	"heracross-mega": "#4E78C2",
	"heracross": "#3C5E7C",
	"herdier": "#986642",
	"hippopotas": "#DFCE8B",
	"hippowdon": "#5B5A4C",
	"hitmonchan": "#C3A7A0",
	"hitmonlee": "#C4AA91",
	"hitmontop": "#57534E",
	"honchkrow": "#446A88",
	"honedge": "#525758",
	"ho-oh": "#6E5A45",
	"hoopa": "#5D5048",
	"hoopa-unbound": "#5C4B4C",
	"hoothoot": "#55463F",
	"hoppip": "#ED7C9D",
	"horsea": "#71C2F4",
	"houndoom-mega": "#4B4343",
	"houndoom": "#8D7D70",
	"houndour": "#89776B",
	"houndstone": "#D4D5DB",
	"huntail": "#78A9D2",
	"hydrapple": "#322924",
	"hydreigon": "#514B50",
	"hypno": "#E6D148",
	"igglybuff": "#F7D7EE",
	"illumise": "#A3ADDB",
	"impidimp": "#E786B1",
	"incineroar": "#724F4E",
	"indeedee": "#534F6A",
	"infernape": "#271D1B",
	"inkay": "#EC9C6B",
	"inteleon-gmax": "#4E545D",
	"inteleon": "#4F5558",
	"iron-boulder": "#2F2C29",
	"iron-bundle": "#8A4746",
	"iron-crown": "#417A81",
	"iron-hands": "#4E5153",
	"iron-jugulis": "#151315",
	"iron-leaves": "#4E7950",
	"iron-moth": "#564438",
	"iron-thorns": "#5F9D49",
	"iron-treads": "#1A1819",
	"iron-valiant": "#252323",
	"ivysaur": "#40654F",
	"jangmo-o": "#E4D698",
	"jellicent": "#BBE5F4",
	"jigglypuff": "#EBB1A9",
	"jirachi": "#E4DD82",
	"jolteon": "#EDE379",
	"joltik": "#F2DC4F",
	"jumpluff": "#EDE29B",
	"jynx": "#A96361",
	"kabuto": "#AF754E",
	"kabutops": "#BB924B",
	"kadabra": "#60563D",
	"kakuna": "#B0942E",
	"kangaskhan-mega": "#947666",
	"kangaskhan": "#6D5C56",
	"karrablast": "#485764",
	"kartana": "#BD9F65",
	"kecleon": "#788A4A",
	"keldeo": "#242324",
	"keldeo-resolute": "#EA7E54",
	"kilowattrel": "#675D57",
	"kingambit": "#4B3731",
	"kingdra": "#5694BE",
	"kingler-gmax": "#DF9060",
	"kingler": "#EB9658",
	"kirlia": "#80C271",
	"klang": "#B7C0C0",
	"klawf": "#743F33",
	"kleavor-noble": "#BFBB2D",
	"kleavor": "#4F5055",
	"klefki": "#665F5D",
	"klinklang": "#9C9C9B",
	"klink": "#CFD6D6",
	"koffing": "#D7D279",
	"komala": "#605546",
	"kommo-o": "#645C49",
	"koraidon-mount": "#261E1E",
	"koraidon": "#261E1E",
	"krabby": "#CE7B5B",
	"kricketot": "#93775F",
	"kricketune": "#815E4E",
	"krokorok": "#947A62",
	"krookodile": "#5E4B4F",
	"kubfu": "#575652",
	"kyogre": "#745368",
	"kyogre-primal": "#578AC6",
	"kyurem-black": "#565753",
	"kyurem": "#959E96",
	"kyurem-white": "#A5AA9F",
	"lairon": "#D4D9DA",
	"lampent": "#67676C",
	"landorus": "#6D5750",
	"landorus-therian": "#6B524E",
	"lanturn": "#88A6EE",
	"lapras-gmax": "#647A81",
	"lapras": "#5B7384",
	"larvesta": "#F4F3F2",
	"larvitar": "#A6E173",
	"latias-mega": "#9187BA",
	"latias": "#C7725D",
	"latios-mega": "#9186BA",
	"latios": "#548BB8",
	"leafeon": "#606244",
	"leavanny": "#5C7B2A",
	"lechonk": "#5B4F50",
	"ledian": "#62494D",
	"ledyba": "#695246",
	"lickilicky": "#F59AA3",
	"lickitung": "#C26969",
	"liepard": "#AE7096",
	"lileep": "#8C718B",
	"lilligant-hisui-noble": "#C6C92A",
	"lilligant-hisui": "#586946",
	"lilligant": "#B5AB4A",
	"lillipup": "#BE875E",
	"linoone-galar": "#5A5658",
	"linoone": "#ECDDE4",
	"litleo": "#614D48",
	"litten": "#4B4D68",
	"litwick": "#F5F5F7",
	"lokix": "#282727",
	"lombre": "#566544",
	"lopunny-mega": "#A67C5A",
	"lopunny": "#AE8963",
	"lotad": "#4F7F68",
	"loudred": "#ADA6E5",
	"lucario-mega": "#524644",
	"lucario": "#575C5F",
	"ludicolo": "#9CA86A",
	"lugia": "#F3F2F5",
	"lugia-shadow": "#555684",
	"lumineon": "#5F727F",
	"lunala": "#57505A",
	"lunatone": "#D1BD63",
	"lurantis": "#EDADB8",
	"luvdisc": "#EF7C89",
	"luxio": "#555856",
	"luxray": "#555652",
	"lycanroc-dusk": "#605548",
	"lycanroc-midnight": "#B5B8B5",
	"lycanroc": "#AFA399",
	"mabosstiff": "#373535",
	"machamp-gmax": "#635E5A",
	"machamp": "#808886",
	"machoke": "#8C828E",
	"machop": "#A8BDB0",
	"magby": "#D56462",
	"magcargo": "#BE5C3A",
	"magearna-original": "#6A4A3E",
	"magearna": "#AA9B91",
	"magikarp": "#D7874E",
	"magmar": "#A15D47",
	"magmortar": "#DC694C",
	"magnemite": "#484747",
	"magneton": "#87929E",
	"magnezone": "#CCD2D8",
	"makuhita": "#EEDB69",
	"malamar": "#937EB3",
	"mamoswine": "#A27E63",
	"manaphy": "#6CB6E7",
	"mandibuzz": "#554B44",
	"manectric-mega": "#E5D656",
	"manectric": "#E0D97B",
	"mankey": "#8C7049",
	"mantine": "#5A80B0",
	"mantyke": "#5A8BD5",
	"maractus": "#5A6E53",
	"mareanie": "#9C94B1",
	"mareep": "#D5B75F",
	"marill": "#75B3E4",
	"marowak-alola": "#5A5656",
	"marowak": "#7C6B5B",
	"marshadow-gen7": "#667078",
	"marshadow": "#667078",
	"marshtomp": "#80655F",
	"maschiff": "#6A3F30",
	"masquerain": "#C8DFE7",
	"maushold-family-of-four": "#F3F2F4",
	"maushold": "#F4F5F6",
	"mawile-mega": "#7A757C",
	"mawile": "#80756E",
	"medicham-mega": "#977E73",
	"medicham": "#E37E7D",
	"meditite": "#C9C9CA",
	"meganium": "#B2DF79",
	"melmetal-gmax": "#9E9B97",
	"melmetal": "#525152",
	"meloetta-pirouette": "#625754",
	"meloetta": "#9CCB9A",
	"meltan": "#E4CF87",
	"meowscarada": "#335231",
	"meowstic": "#5D7FAD",
	"meowth-alola": "#918E7B",
	"meowth-galar": "#BDAC9D",
	"meowth-gmax": "#F8EBB0",
	"meowth": "#957C46",
	"mesprit": "#605A68",
	"metagross-mega": "#556566",
	"metagross": "#465C6F",
	"metang": "#7097D9",
	"metapod": "#97C251",
	"mew": "#F0B0BA",
	"mewtwo-mega-x": "#9B7F9E",
	"mewtwo-mega-y": "#957B97",
	"mewtwo": "#DDD5E1",
	"mienfoo": "#EFE572",
	"mienshao": "#A093A6",
	"mightyena": "#7D7979",
	"milcery": "#F1ECD3",
	"milotic": "#615362",
	"miltank": "#584F4F",
	"mime-jr": "#6071E5",
	"mimikyu": "#CDC18F",
	"minccino": "#D2D0D1",
	"minior-blue-gen7": "#77C1EA",
	"minior-blue": "#77C1EA",
	"minior-green-gen7": "#8CCC56",
	"minior-green": "#8CCC56",
	"minior-indigo-gen7": "#5992CC",
	"minior-indigo": "#5992CC",
	"minior-orange-gen7": "#ECA35A",
	"minior-orange": "#ECA35A",
	"minior": "#87655B",
	"minior-red-gen7": "#DB7095",
	"minior-red": "#DB7095",
	"minior-violet-gen7": "#9887D1",
	"minior-violet": "#9887D1",
	"minior-yellow-gen7": "#E8D24A",
	"minior-yellow": "#E8D24A",
	"minun": "#E4DF6B",
	"miraidon-mount": "#1C1C26",
	"miraidon": "#1C1C26",
	"misdreavus": "#598AAC",
	"mismagius": "#A170A1",
	"moltres-galar": "#5A5758",
	"moltres": "#EEA150",
	"monferno": "#E6AB60",
	"morelull": "#F1DDDC",
	"morgrem": "#523F4F",
	"morpeko-hangry": "#735C76",
	"morpeko": "#736757",
	"mothim": "#7F6A50",
	"mr-mime-galar": "#4B6278",
	"mr-mime": "#4F5874",
	"mr-rime": "#616386",
	"mudbray": "#A66E59",
	"mudkip": "#8CCCEE",
	"mudsdale": "#5F6369",
	"muk-alola": "#74889D",
	"muk": "#AD94D0",
	"munchlax": "#5D8CA2",
	"munkidori": "#3C3A40",
	"munna": "#FCBDD0",
	"murkrow": "#515F8C",
	"musharna": "#F396A8",
	"nacli": "#AF8B7B",
	"naclstack": "#64483D",
	"naganadel": "#8C5E9B",
	"natu": "#656E2B",
	"necrozma-dawn": "#504F4E",
	"necrozma-dusk": "#524F47",
	"necrozma": "#5A5959",
	"necrozma-ultra": "#C9B776",
	"nickit": "#4E423F",
	"nidoking": "#896E97",
	"nidoqueen": "#636D78",
	"nidoran-f": "#7B8AB5",
	"nidoran-m": "#B27EA2",
	"nidorina": "#6C908F",
	"nidorino": "#B580A5",
	"nihilego": "#8DADD0",
	"nincada": "#A0A594",
	"ninetales-alola": "#BBDDF2",
	"ninetales": "#C1AC77",
	"ninjask": "#5C524C",
	"noctowl": "#584943",
	"noibat": "#967AB1",
	"noivern": "#574A4E",
	"nosepass": "#7778A1",
	"numel": "#C2C05C",
	"nuzleaf": "#AA8C6F",
	"nymble": "#7B8790",
	"obstagoon": "#504D50",
	"octillery": "#C13E3D",
	"oddish": "#425740",
	"ogerpon-cornerstone-mask": "#364133",
	"ogerpon-hearthflame-mask": "#43482F",
	"ogerpon-teal-mask": "#2B502D",
	"ogerpon-wellspring-mask": "#192C28",
	"oinkologne-f": "#5C3E3E",
	"oinkologne": "#4C4858",
	"okidogi": "#3C3937",
	"omanyte": "#E7DBA1",
	"omastar": "#5A6962",
	"onix": "#767776",
	"oranguru": "#6F7472",
	"orbeetle-gmax": "#6D4F53",
	"orbeetle": "#564850",
	"oricorio-pau": "#DFB5AF",
	"oricorio": "#5E444D",
	"oricorio-pom-pom": "#EFE390",
	"oricorio-sensu": "#9EA4D4",
	"orthworm": "#724D4F",
	"oshawott": "#ECEFE8",
	"overqwil": "#645E60",
	"pachirisu": "#EFF3F1",
	"palafin-hero": "#4783B1",
	"palafin": "#50C3D8",
	"palkia-origin": "#A69098",
	"palkia": "#605559",
	"palossand": "#F8C183",
	"palpitoad": "#545B64",
	"pancham": "#434840",
	"pangoro": "#505254",
	"panpour": "#7CBAEC",
	"pansage": "#41C05C",
	"pansear": "#E5624D",
	"parasect": "#EC935E",
	"paras": "#D18143",
	"passimian": "#585D4F",
	"patrat": "#C38C52",
	"pawmi": "#855632",
	"pawmo": "#ECB35C",
	"pawmot": "#D99819",
	"pawniard": "#5F5454",
	"pecharunt": "#CF599F",
	"pelipper": "#F3DD48",
	"perrserker": "#585453",
	"persian-alola": "#838895",
	"persian": "#F7EDB5",
	"petilil": "#90C051",
	"phanpy": "#88C7E8",
	"phantump": "#4B4B3E",
	"pheromosa": "#E4E2DC",
	"phione": "#69B3EC",
	"pichu": "#15130F",
	"pichu-spiky-eared": "#F0E569",
	"pidgeot-mega": "#B35359",
	"pidgeot": "#CC7951",
	"pidgeotto": "#BD6C40",
	"pidgey": "#BA8646",
	"pidove": "#B9ACA4",
	"pignite": "#E69A44",
	"pikachu-alola-cap": "#DFA642",
	"pikachu-belle": "#292822",
	"pikachu-cosplay": "#F2D233",
	"pikachu-gmax": "#EDD742",
	"pikachu-hoenn-cap": "#B0764C",
	"pikachu-kalos-cap": "#E2AE37",
	"pikachu-libre": "#D2A946",
	"pikachu-original-cap": "#E1C131",
	"pikachu-partner-cap": "#E2C130",
	"pikachu-phd": "#5D5846",
	"pikachu": "#E3C031",
	"pikachu-pop-star": "#D876A0",
	"pikachu-rock-star": "#734B37",
	"pikachu-sinnoh-cap": "#E7C631",
	"pikachu-starter": "#DBB52E",
	"pikachu-unova-cap": "#E0B036",
	"pikachu-world-cap": "#C67B3B",
	"pikipek": "#565151",
	"piloswine": "#F2AF81",
	"pincurchin": "#514756",
	"pineco": "#42615C",
	"pinsir-mega": "#71614F",
	"pinsir": "#98887D",
	"piplup": "#599CE8",
	"plusle": "#F1E871",
	"poipole": "#5A4768",
	"politoed": "#85C438",
	"poliwag": "#778CBB",
	"poliwhirl": "#7996C6",
	"poliwrath": "#557092",
	"poltchageist": "#323631",
	"polteageist": "#615354",
	"ponyta-galar": "#E2D9EC",
	"ponyta": "#DE9B40",
	"poochyena": "#5D5959",
	"popplio": "#6089D0",
	"porygon2": "#D95D7A",
	"porygon": "#62AAE1",
	"porygon-z": "#DC7262",
	"primarina": "#CFD5E4",
	"primeape": "#E5CC94",
	"prinplup": "#4975A5",
	"probopass": "#D05552",
	"psyduck": "#DDC668",
	"pumpkaboo-large": "#7B5738",
	"pumpkaboo": "#7A5738",
	"pumpkaboo-small": "#7B5837",
	"pumpkaboo-super": "#7A5636",
	"pupitar": "#B3B9DE",
	"purrloin": "#8969AC",
	"purugly": "#9190AF",
	"pyroar": "#775241",
	"pyukumuku": "#534949",
	"quagsire": "#7DD3FB",
	"quaquaval": "#3154B1",
	"quaxly": "#D7DBCC",
	"quaxwell": "#20232D",
	"quilava": "#1F120C",
	"quilladin": "#69683F",
	"qwilfish-hisui": "#625E5F",
	"qwilfish": "#55A39B",
	"raboot": "#604F43",
	"rabsca": "#B8328E",
	"raging-bolt": "#6E6E86",
	"raichu-alola": "#E8CA5C",
	"raichu": "#D8AA46",
	"raikou": "#746769",
	"ralts": "#7DB770",
	"rampardos": "#595B61",
	"rapidash-galar": "#A3A1CA",
	"rapidash": "#F2E697",
	"raticate-alola": "#54504D",
	"raticate": "#B18E53",
	"rattata-alola": "#6C6B6A",
	"rattata": "#AB93B6",
	"rayquaza-mega": "#5F7658",
	"rayquaza": "#4B5D46",
	"regice": "#6287BC",
	"regidrago": "#3B4C48",
	"regieleki": "#EEED64",
	"regigigas": "#5A6653",
	"regirock": "#BB8466",
	"registeel": "#898686",
	"relicanth": "#6F6154",
	"rellor": "#936F4C",
	"remoraid": "#B0DAEB",
	"reshiram": "#ADB0D5",
	"reuniclus": "#8EB693",
	"revavroom": "#494749",
	"rhydon": "#C6BFBD",
	"rhyhorn": "#727E85",
	"rhyperior": "#6B564E",
	"ribombee": "#F2F1F0",
	"rillaboom-gmax": "#446448",
	"rillaboom": "#415244",
	"riolu": "#424D5D",
	"roaring-moon": "#181822",
	"rockruff": "#595044",
	"roggenrola": "#5668A6",
	"rolycoly": "#4B4944",
	"rookidee": "#45464C",
	"roselia": "#609968",
	"roserade": "#57724F",
	"rotom-fan": "#66583B",
	"rotom-frost": "#E17140",
	"rotom-heat": "#EB6A4C",
	"rotom-mow": "#53583D",
	"rotom": "#76CEF6",
	"rotom-wash": "#E17242",
	"rowlet": "#CAA58D",
	"rufflet": "#ECE7C3",
	"runerigus": "#AB5464",
	"sableye-mega": "#D2375B",
	"sableye": "#796DA0",
	"salamence-mega": "#629DC0",
	"salamence": "#6AACB4",
	"salandit": "#443E3E",
	"salazzle": "#746679",
	"samurott-hisui": "#41486B",
	"samurott": "#516372",
	"sandaconda-gmax": "#584D44",
	"sandaconda": "#484137",
	"sandile": "#6F6153",
	"sandshrew-alola": "#DEEBEB",
	"sandshrew": "#A7953E",
	"sandslash-alola": "#608EAC",
	"sandslash": "#5E4A3A",
	"sandygast": "#DEDDAC",
	"sandy-shocks": "#494949",
	"sawk": "#404951",
	"sawsbuck-autumn": "#CB6452",
	"sawsbuck": "#A76E56",
	"sawsbuck-summer": "#C28D62",
	"sawsbuck-winter": "#AA8A78",
	"scatterbug": "#7C7974",
	"sceptile-mega": "#436443",
	"sceptile": "#6D8968",
	"scizor-mega": "#5B4A49",
	"scizor": "#74454A",
	"scolipede": "#7A525D",
	"scorbunny": "#C19C75",
	"scovillain": "#42483A",
	"scrafty": "#DA914D",
	"scraggy": "#E9AB32",
	"scream-tail": "#C29289",
	"scyther": "#4D6F54",
	"seadra": "#65B2ED",
	"seaking": "#D18979",
	"sealeo": "#56ABD8",
	"seedot": "#4A4740",
	"seel": "#E7EAF4",
	"seismitoad": "#4D5256",
	"sentret": "#CF9A89",
	"serperior": "#60A353",
	"servine": "#7BC14B",
	"seviper": "#565253",
	"sewaddle": "#689836",
	"sharpedo-mega": "#5077BA",
	"sharpedo": "#4F77A0",
	"shaymin": "#527342",
	"shaymin-sky": "#7D8250",
	"shedinja": "#A49054",
	"shelgon": "#ACABA5",
	"shellder": "#494855",
	"shellos-east": "#71BEE8",
	"shellos": "#DB8A83",
	"shelmet": "#8A878B",
	"shieldon": "#7C7970",
	"shiftry": "#756F5F",
	"shiinotic": "#BA7FBE",
	"shinx": "#7AB8EA",
	"shroodle": "#45495C",
	"shroomish": "#E6CF82",
	"shuckle": "#8C5140",
	"shuppet": "#7F89B7",
	"sigilyph": "#4D4F43",
	"silcoon": "#EDF1F3",
	"silicobra": "#585340",
	"silvally-bug": "#515354",
	"silvally-dark": "#505156",
	"silvally-dragon": "#4D5358",
	"silvally-electric": "#565756",
	"silvally-fairy": "#515056",
	"silvally-fighting": "#555454",
	"silvally-fire": "#7B4D46",
	"silvally-flying": "#55585F",
	"silvally-ghost": "#535058",
	"silvally-grass": "#3C6F59",
	"silvally-ground": "#635546",
	"silvally-ice": "#505158",
	"silvally": "#525257",
	"silvally-poison": "#514E5A",
	"silvally-psychic": "#524E57",
	"silvally-rock": "#525257",
	"silvally-steel": "#505258",
	"silvally-water": "#4D5157",
	"simipour": "#7DBAEA",
	"simisage": "#41B057",
	"simisear": "#E4674E",
	"sinistcha": "#393C32",
	"sinistea": "#7D678D",
	"sirfetchd": "#3D644B",
	"sizzlipede": "#683527",
	"skarmory": "#CED0E6",
	"skeledirge": "#784843",
	"skiddo": "#486733",
	"skiploom": "#B1DE43",
	"skitty": "#F2BC92",
	"skorupi": "#7E8BD7",
	"skrelp": "#A25A45",
	"skuntank": "#7D6D9D",
	"skwovet": "#827760",
	"slaking": "#B8988A",
	"slakoth": "#E5CEB0",
	"sliggoo-hisui": "#968BA5",
	"sliggoo": "#C5AEE9",
	"slither-wing": "#36262B",
	"slowbro-galar": "#A088A3",
	"slowbro-mega": "#A48F7F",
	"slowbro": "#F1C7CC",
	"slowking-galar": "#BFA2BF",
	"slowking": "#B99992",
	"slowpoke-galar": "#A8748B",
	"slowpoke": "#C7867B",
	"slugma": "#CA5E3D",
	"slurpuff": "#F8D7E2",
	"smeargle": "#EFEBBB",
	"smoliv": "#8D9748",
	"smoochum": "#EFE98C",
	"sneasel-hisui": "#8571A9",
	"sneasel": "#5D7D96",
	"sneasler": "#5A545E",
	"snivy": "#6DB145",
	"snom": "#F2F2F9",
	"snorlax-gmax": "#227067",
	"snorlax": "#40637E",
	"snorunt": "#504A44",
	"snover": "#F1F1F0",
	"snubbull": "#E08BA0",
	"sobble": "#AFD6EB",
	"solgaleo": "#332E28",
	"solosis": "#67A971",
	"solrock": "#ECB33B",
	"spearow": "#B76E3D",
	"spectrier": "#4A4654",
	"spewpa": "#645553",
	"spheal": "#83A4E5",
	"spidops": "#383829",
	"spinarak": "#6B6949",
	"spinda-blank": "#AC8A7E",
	"spinda-filled": "#955654",
	"spinda": "#B67770",
	"spiritomb": "#D19FBF",
	"spoink": "#D4BEBB",
	"sprigatito": "#99B374",
	"spritzee": "#E07990",
	"squawkabilly-blue": "#151B25",
	"squawkabilly-green": "#4E6642",
	"squawkabilly-white": "#5B5650",
	"squawkabilly-yellow": "#685D43",
	"squirtle": "#4A6F70",
	"stakataka": "#686C6D",
	"stantler": "#DDB588",
	"staraptor": "#574E46",
	"staravia": "#5A574F",
	"starly": "#57534C",
	"starmie": "#665356",
	"staryu": "#A57F51",
	"steelix-mega": "#708092",
	"steelix": "#9296AA",
	"steenee": "#AEB779",
	"stonjourner": "#847C73",
	"stoutland": "#6A6563",
	"stufful": "#6E564E",
	"stunfisk-galar": "#565453",
	"stunfisk": "#D8A569",
	"stunky": "#82719E",
	"sudowoodo": "#D39561",
	"suicune": "#7287A0",
	"sunflora": "#96A449",
	"sunkern": "#5F652D",
	"surskit": "#90BEE7",
	"swablu": "#84ADD2",
	"swadloon": "#99D940",
	"swalot": "#A78FCB",
	"swampert-mega": "#525B60",
	"swampert": "#5D6B7B",
	"swanna": "#F3F5F4",
	"swellow": "#6D69B4",
	"swinub": "#ECA375",
	"swirlix": "#F8F0F6",
	"swoobat": "#6A6C70",
	"sylveon": "#EBC3D1",
	"tadbulb": "#181716",
	"taillow": "#63609F",
	"talonflame": "#625547",
	"tandemaus": "#F4F5F6",
	"tangela": "#474A54",
	"tangrowth": "#4D4D5F",
	"tapu-bulu": "#584C3D",
	"tapu-fini": "#585658",
	"tapu-koko": "#ECBC5F",
	"tapu-lele": "#5E4F4B",
	"tarountula": "#D2D1CC",
	"tatsugiri-droopy": "#E54071",
	"tatsugiri": "#DF754B",
	"tatsugiri-stretchy": "#F5DF0C",
	"tauros-aqua-breed": "#3B3C3D",
	"tauros-blaze-breed": "#3F3C3C",
	"tauros-combat-breed": "#141313",
	"tauros": "#C69C6E",
	"teddiursa": "#DD9132",
	"tentacool": "#6BA0D8",
	"tentacruel": "#A06D6C",
	"tepig": "#6F553B",
	"terapagos": "#373D5E",
	"terapagos-stellar": "#2742A0",
	"terapagos-terastal": "#323A62",
	"terrakion": "#64544C",
	"thievul": "#564946",
	"throh": "#504742",
	"thundurus": "#535A62",
	"thundurus-therian": "#494E5B",
	"thwackey": "#5A5841",
	"timburr": "#6A554A",
	"ting-lu": "#3F4A44",
	"tinkatink": "#D85183",
	"tinkaton": "#916892",
	"tinkatuff": "#5D4162",
	"tirtouga": "#626A89",
	"toedscool": "#C9B964",
	"toedscruel": "#6D644A",
	"togedemaru": "#DEDDDB",
	"togekiss": "#DDF3E9",
	"togepi": "#E0E4CC",
	"togetic": "#F1F1F1",
	"torchic": "#F1A82F",
	"torkoal": "#6B584D",
	"tornadus": "#657059",
	"tornadus-therian": "#5A535A",
	"torracat": "#604B46",
	"torterra": "#48693D",
	"totodile": "#7DC5E6",
	"toucannon": "#545451",
	"toxapex": "#8A849D",
	"toxel": "#BA92DC",
	"toxicroak": "#708CD3",
	"toxtricity-gmax": "#EDC0BB",
	"toxtricity-low-key": "#786CB3",
	"toxtricity": "#5A4D51",
	"tranquill": "#5D5958",
	"trapinch": "#EA9177",
	"treecko": "#4D763A",
	"trevenant": "#565B4D",
	"tropius": "#56844F",
	"trubbish": "#7D9E67",
	"trumbeak": "#4A4C47",
	"tsareena": "#778A5A",
	"turtonator": "#744D4B",
	"turtwig": "#C3D46D",
	"tympole": "#CFC1A5",
	"tynamo": "#D5E5F2",
	"type-null": "#54504F",
	"typhlosion-hisui": "#E8D579",
	"typhlosion": "#EAD778",
	"tyranitar-mega": "#5C653A",
	"tyranitar": "#B2CB95",
	"tyrantrum": "#704042",
	"tyrogue": "#D5B8D5",
	"tyrunt": "#705855",
	"umbreon": "#646D73",
	"unfezant": "#5D5A57",
	"unown-b": "#1B1B1B",
	"unown-c": "#494949",
	"unown-d": "#1E1E1E",
	"unown-e": "#898989",
	"unown-exclamation": "#8F8F8F",
	"unown-f": "#4E4E4E",
	"unown-g": "#737373",
	"unown-h": "#737373",
	"unown-i": "#1F1F1F",
	"unown-j": "#4A4A4A",
	"unown-k": "#474747",
	"unown-l": "#4D4D4D",
	"unown-m": "#777777",
	"unown-n": "#1E1E1E",
	"unown-o": "#1D1D1D",
	"unown": "#424242",
	"unown-p": "#1E1E1E",
	"unown-q": "#848484",
	"unown-question": "#959595",
	"unown-r": "#1C1C1C",
	"unown-s": "#1C1C1C",
	"unown-t": "#444444",
	"unown-u": "#1B1B1B",
	"unown-v": "#707070",
	"unown-w": "#868686",
	"unown-x": "#212121",
	"unown-y": "#232323",
	"unown-z": "#4C4C4C",
	"ursaluna": "#60524C",
	"ursaluna-bloodmoon": "#9B7761",
	"ursaring": "#BE8958",
	"urshifu-gmax": "#61343C",
	"urshifu": "#515356",
	"urshifu-rapid-strike-gmax": "#42455C",
	"uxie": "#66655A",
	"vanillish": "#E8F5FB",
	"vanillite": "#E5F5FB",
	"vanilluxe": "#86AFD0",
	"vaporeon": "#587281",
	"varoom": "#414041",
	"veluza": "#5F495B",
	"venipede": "#54483F",
	"venomoth": "#D2C7D9",
	"venonat": "#8178C4",
	"venusaur-gmax": "#D16F66",
	"venusaur-mega": "#555652",
	"venusaur": "#70B8A2",
	"vespiquen": "#6E5B45",
	"vibrava": "#6D9B63",
	"victini": "#CF8053",
	"victreebel": "#78B76E",
	"vigoroth": "#E4E2E2",
	"vikavolt": "#506375",
	"vileplume": "#E5686D",
	"virizion": "#767F52",
	"vivillon-archipelago": "#4B413C",
	"vivillon-continental": "#937156",
	"vivillon-elegant": "#6C5B75",
	"vivillon-fancy": "#8E9E74",
	"vivillon-garden": "#405C45",
	"vivillon-high-plains": "#62554C",
	"vivillon-icy-snow": "#464646",
	"vivillon-jungle": "#47574B",
	"vivillon-marine": "#4E5461",
	"vivillon-modern": "#4F4847",
	"vivillon-monsoon": "#606164",
	"vivillon-ocean": "#4E4946",
	"vivillon": "#9D5B68",
	"vivillon-poke-ball": "#8C605A",
	"vivillon-polar": "#3D3F4B",
	"vivillon-river": "#565C59",
	"vivillon-sandstorm": "#635E53",
	"vivillon-savanna": "#546876",
	"vivillon-sun": "#CE875A",
	"vivillon-tundra": "#607C8F",
	"volbeat": "#5D4C47",
	"volcanion": "#794145",
	"volcarona": "#6D5851",
	"voltorb-hisui": "#D56C0F",
	"voltorb": "#CD6045",
	"vullaby": "#79635D",
	"vulpix-alola": "#F4F6F8",
	"vulpix": "#A8663F",
	"wailmer": "#5599C8",
	"wailord": "#56B4FA",
	"walking-wake": "#455568",
	"walrein": "#6D98C4",
	"wartortle": "#5C5A5C",
	"watchog": "#D08547",
	"wattrel": "#5D585A",
	"weavile": "#604655",
	"weedle": "#C7914D",
	"weepinbell": "#7FA85A",
	"weezing-galar": "#8F9684",
	"weezing": "#DAD191",
	"whimsicott": "#F5F4DB",
	"whirlipede": "#7A6C76",
	"whiscash": "#4F61A1",
	"whismur": "#DEB6D9",
	"wigglytuff": "#F6DEDE",
	"wiglett": "#9EA5A0",
	"wimpod": "#5C5363",
	"wingull": "#EFF3F5",
	"wishiwashi": "#535F7C",
	"wishiwashi-school": "#3C4D6E",
	"wobbuffet": "#77C5EC",
	"wo-chien": "#564F45",
	"woobat": "#6F6C78",
	"wooloo": "#F3EFEF",
	"wooper-paldea": "#664B49",
	"wooper": "#96B7E6",
	"wormadam": "#6B8561",
	"wormadam-sandy": "#625B4F",
	"wormadam-trash": "#8F6C6F",
	"wugtrio": "#5E4E53",
	"wurmple": "#E86767",
	"wynaut": "#6EC3EE",
	"wyrdeer": "#D8D7D7",
	"xatu": "#3D372F",
	"xerneas-active": "#585E6A",
	"xerneas": "#38424E",
	"xurkitree": "#524F47",
	"yamask-galar": "#51494B",
	"yamask": "#6C5951",
	"yamper": "#C4BA69",
	"yanma": "#604E3B",
	"yanmega": "#968250",
	"yungoos": "#977355",
	"yveltal": "#6D6161",
	"zacian-crowned": "#585B5C",
	"zacian": "#5B5155",
	"zamazenta-crowned": "#8B6B71",
	"zamazenta": "#554C6E",
	"zangoose": "#C09291",
	"zapdos-galar": "#4C463F",
	"zapdos": "#F1DD49",
	"zarude-dada": "#545552",
	"zarude": "#535652",
	"zebstrika": "#707471",
	"zekrom": "#454749",
	"zeraora": "#646645",
	"zigzagoon-galar": "#514F50",
	"zigzagoon": "#98846D",
	"zoroark-hisui": "#614B4D",
	"zoroark": "#57474D",
	"zorua-hisui": "#B28584",
	"zorua": "#63505B",
	"zubat": "#4C707C",
	"zweilous": "#716469",
	"zygarde-10": "#5A5E4D",
	"zygarde-complete": "#404442",
	"zygarde": "#70755E",
}



import base64

# Logos
logo_black_path = './assets/logo_black.png'
logo_black_tunel = base64.b64encode(open(logo_black_path, 'rb').read())
logo_white_path = './assets/logo.png'
logo_white_tunel = base64.b64encode(open(logo_white_path, 'rb').read())

logo_black_sm_path = './assets/logo_black_small.png'
logo_black_sm_tunel = base64.b64encode(open(logo_black_sm_path, 'rb').read())
logo_white_sm_path = './assets/logo_small.png'
logo_white_sm_tunel = base64.b64encode(open(logo_white_sm_path, 'rb').read())



import re

red = '#e74c3c'
white = '#ffffff'
blue = '#3498db'
green = '#18bc9c'
primary = '#2c3e50'


def hex_to_rgb(hex):
    """ Convert a hex value to list of rgb triplets """
    return [int(hex[i:i+2], 16) for i in range(1, 6, 2)]


def rgb_to_hex(rgb):
    """ Convert a list of rgb triplets to a hex list """
    if isinstance(rgb, str) and rgb.startswith('rgb'):
        matches = re.findall(r'\d+', rgb)
        if matches:
            # Convert the extracted numbers to integers
            rgb = tuple(map(int, matches))
    RGB = [int(x) for x in rgb]
    return '#' + ''.join(['0{0:x}'.format(v) if v < 16 else '{0:x}'.format(v) for v in RGB])


def linear_gradient(start_hex, finish_hex, n=50):
    """ Create a gradient of n values between 2 initial hex values """
    s = hex_to_rgb(start_hex)
    f = hex_to_rgb(finish_hex)

    rgb_list = [start_hex]
    for t in range(1, n):
        curr_vector = [
            int(s[j] + (float(t)/(n-1))*(f[j]-s[j]))
            for j in range(3)
        ]
        rgb_list.append(rgb_to_hex(curr_vector))

    return rgb_list


def transparent_gradient(hex_color, n=50):
    # Convert hex to RGB
    hex_color = hex_color.lstrip('#')
    r, g, b = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))

    # Create gradient
    gradient = []
    for i in range(n):
        alpha = i / (n - 1)
        rgba = (r, g, b, alpha)
        gradient.append(f"rgba({rgba[0]}, {rgba[1]}, {rgba[2]}, {rgba[3]:.2f})")

    return gradient


def create_color_map(color_list):
    """ Create the color map for graph objects """
    total_colors = len(color_list)
    colormap = [[0, color_list[0]]]
    for i, item in enumerate(color_list):
        ratio = i / total_colors
        colormap.append([ratio, item])
    colormap.append([1, color_list[-1]])
    return colormap


red_to_white = linear_gradient(red, white)
white_to_blue = linear_gradient(white, blue)
red_to_white_to_blue = red_to_white + white_to_blue

win_rate_color_bar = create_color_map(red_to_white_to_blue)

green_gradient = transparent_gradient(green, 101)
red_gradient = transparent_gradient(red, 101)
blue_gradient = transparent_gradient(blue, 101)


def text_color_for_background(rgb):
    rgb = rgb if not isinstance(rgb, str) and rgb.startswith('#') else hex_to_rgb(rgb)
    r, g, b = rgb

    def convert(color):
        color /= 255.0
        if color <= 0.03928:
            return color / 12.92
        return ((color + 0.055) / 1.055) ** 2.4

    R = convert(r)
    G = convert(g)
    B = convert(b)

    Y = 0.2126 * R + 0.7152 * G + 0.0722 * B
    return 'black' if Y > 0.5 else 'white'



from datetime import timedelta, date

def days_ago(days):
    today = date.today()
    days_ago_str = str(today - timedelta(days))
    return days_ago_str


def weeks_ago_3():
    today = date.today()
    weeks_ago_3 = str(today - timedelta(21))
    return weeks_ago_3
